# tests/test_pressure_unit_conversion.py
import pandas as pd
import pytest

from neptoon.io.read.data_ingest import (
    InputDataFrameFormattingConfig,
    FormatDataForCRNSDataHub,
    InputColumnMetaData,
    InputColumnDataType,
    PressureUnits,
)


@pytest.fixture
def base_config():
    cfg = InputDataFrameFormattingConfig(path_to_config=None)
    return cfg


def test_pascals_to_hectopascals(base_config):
    df = pd.DataFrame({"P_raw": [101325, 100000, 98000]})

    base_config.column_data = [
        InputColumnMetaData(
            initial_name="P_raw",
            variable_type=InputColumnDataType.PRESSURE,
            unit=PressureUnits.PASCALS,
            priority=1,
        )
    ]
    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.standardise_units_of_pressure()
    out = formatter.data_frame

    expected = df["P_raw"] / 100
    pd.testing.assert_series_equal(out["P_raw"], expected, check_names=False)

    assert base_config.column_data[0].unit == PressureUnits.HECTOPASCALS


def test_kilopascals_to_hectopascals(base_config):
    df = pd.DataFrame({"P_kpa": [101.325, 100.0, 98.0]})

    base_config.column_data = [
        InputColumnMetaData(
            initial_name="P_kpa",
            variable_type=InputColumnDataType.PRESSURE,
            unit=PressureUnits.KILOPASCALS,
            priority=1,
        )
    ]

    formatter = FormatDataForCRNSDataHub(
        data_frame=df.copy(), config=base_config
    )
    formatter.standardise_units_of_pressure()
    out = formatter.data_frame

    expected = df["P_kpa"] * 10
    pd.testing.assert_series_equal(out["P_kpa"], expected, check_names=False)

    assert base_config.column_data[0].unit == PressureUnits.HECTOPASCALS



import pandas as pd
from pandera.errors import SchemaError
import pytest

from neptoon.quality_control.data_validation_tables import (
    RawDataSchemaAfterFirstQA,
)


def test_relative_humidity_check():
    """
    Checks if the check on relative humidity is working as expected. It
    should raise an error if the RH values are in decimal format.
    """

    df_invalid = pd.DataFrame(
        {
            "moderated_count": [2000, 2010, 2001, 1980, 1999],
            "atmos_pressure": [990, 1000, 999, 956, 1110],
            # "relative_humidity": [25, 27, 13, 14, 15],
            "air_relative_humidity": [0.2, 0.3, 0.3, 0.4, 0.5],
            "date_time": ["2001", "2001", "2001", "2001", "2001"],
            "air_temperature": [10, 10, 10, 10, 10],
        }
    )

    with pytest.raises(SchemaError):
        RawDataSchemaAfterFirstQA.validate(df_invalid)



import requests
from datetime import datetime
import logging
import pandas as pd
from pathlib import Path
from io import StringIO
from dateutil import parser

from neptoon.columns import ColumnInfo
from neptoon.config.global_configuration import GlobalConfig
from neptoon.data_audit import log_key_step
from neptoon.logging import get_logger

core_logger = get_logger()

NMDB_REFERENCES = {
    "JUNG": 161,
    "SOPO": 308,
    "OULU": 108,
    "PSNM": 615,
    "MXCO": 227,
    "AATA": 157,
    "INVK": 111,
    "KIEL": 166,
}

NMDB_CUTOFF_RIGIDITIES = {
    "JUNG": 4.49,
    "SOPO": 0.0,
    "OULU": 0.619,
    "PSNM": 16.674,
    "MXCO": 7.495,
    "AATA": 6.018,
    "INVK": 0.186,
    "KIEL": 2.383,
}


class NMDBDataAttacher:
    """
    This is the core class that a user interacts with when wanting to
    attach data from the NMDB.eu database to a dataframe. It includes
    methods for configuring the NMDBConfig class which is then used by
    other classes for fetching and parsing data from the NMDB.eu API.

    TODO - add validation steps to ensure dataframe is correct format
    """

    def __init__(
        self,
        data_frame: pd.DataFrame,
        new_column_name=str(ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY),
    ):
        """
        Initialisation parameters

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame which requires data to be attached. It must have a
            datetime index.
        new_column_name : str, optional
            column name for the new column were neutron count data is
            appended, by default "incoming_neutron_intensity"
        """
        self.data_frame = data_frame
        self._new_column_name = new_column_name

    @property
    def new_column_name(self):
        return self._new_column_name

    def configure(
        self,
        station: str,
        reference_value: int | None = None,
        resolution="60",
        nmdb_table="revori",
    ):
        start_date_from_data = self.data_frame.index[0]
        end_date_from_data = self.data_frame.index[-1]
        self.config = NMDBConfig(
            start_date_wanted=start_date_from_data,
            end_date_wanted=end_date_from_data,
            station=station,
            reference_value=reference_value,
            resolution=resolution,
            nmdb_table=nmdb_table,
        )

    def fetch_data(self):
        """
        Creates a NMDBDataHandler using the config and collects the data
        whilst storing it under self.tmp_data
        """
        handler = NMDBDataHandler(self.config)
        self.tmp_data = handler.collect_nmdb_data()

    def attach_data(self):
        """
        Attaches the data stored in self.tmp_data to self.data_frame.
        This occurs inplace.

        Raises
        ------
        ValueError
            When index of the data is not Datetime an error occurs
        """
        if self.config.reference_value is None:
            if self.config.station not in list(NMDB_REFERENCES.keys()):
                message = (
                    "NMDB station not supported for automatic "
                    "reference creation. Please choose one of: \n"
                    f"{list(NMDB_REFERENCES.keys())}"
                )
                core_logger.error(message)
                raise ValueError(message)
            else:
                self.config.reference_value = NMDB_REFERENCES[
                    self.config.station
                ]

        if not isinstance(self.tmp_data.index, pd.DatetimeIndex):
            raise ValueError("DataFrame source must have a DatetimeIndex.")

        mapped_data = self.tmp_data["count"].reindex(
            self.data_frame.index, method="nearest"
        )
        self.data_frame[self.new_column_name] = mapped_data
        self.data_frame[
            str(ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE)
        ] = self.config.reference_value

        self.data_frame[str(ColumnInfo.Name.NMDB_REFERENCE_STATION)] = (
            self.config.station
        )
        self.data_frame[
            str(ColumnInfo.Name.REFERENCE_MONITOR_CUTOFF_RIGIDITY)
        ] = NMDB_CUTOFF_RIGIDITIES[self.config.station]

    def return_data_frame(self):
        """
        Returns the DataFrame attached in the object.

        Returns
        -------
        pd.DataFrame
            The DataFrame
        """
        return self.data_frame


class DateTimeHandler:
    """
    Class that holds Date standardization methods.

    This class provides static methods for converting and standardizing
    date formats to a common format (YYYY-mm-dd) used throughout the
    NMDB data collection process.
    """

    @staticmethod
    def convert_string_to_standard_date(date_str):
        """
        Converts a string date to the standard format (YYYY-mm-dd).

        Parameters
        ----------
        date_str : str
            The date string to convert.

        Returns
        -------
        str or None
            The standardized date string in format (YYYY-mm-dd), or None
            if the input date string is not a recognizable date format.

        Raises
        ------
        ValueError
            If the input string cannot be parsed into a valid date.
        """
        try:
            parsed_date = parser.parse(date_str)
            standardized_date_str = parsed_date.strftime("%Y-%m-%d")
            return standardized_date_str

        except ValueError:
            logging.error(
                f"Error: '{date_str}' is not a recognizable date format."
            )
            return None

    @staticmethod
    def format_datetime_to_standard_string(date_datetime):
        """
        Converts a datetime or pd.Timestamp date to the standard format
        (YYYY-mm-dd)

        Parameters
        ----------
        date_datetime : datetime.datetime or pandas.Timestamp
            The datetime object to convert.

        Returns
        -------
        str or None The standardized date string in format (YYYY-mm-dd),
        or None if the input
            is not a datetime.datetime or pandas.Timestamp instance.

        """
        if isinstance(date_datetime, (datetime, pd.Timestamp)):
            return date_datetime.strftime("%Y-%m-%d")
        else:
            logging.error("Input is not a valid datetime object")
            return None

    @staticmethod
    def standardize_date_input(date_input):
        """
        Takes a date as input, checks type, and converts it to the
        standard format (YYYY-mm-dd)

        Parameters
        ----------
        date_input : str or pd.TimeStamp or datetime.datetime
            Input date to be converted

        Returns
        -------
        str
            String of the date as YYYY-mm-dd

        Raises
        ------
        ValueError
            Raise error when neither str or datetime is given
        """
        if isinstance(date_input, (datetime, pd.Timestamp)):
            logging.info("Date was given as a pandas.Timestamp")
            return DateTimeHandler.format_datetime_to_standard_string(
                date_input
            )
        elif isinstance(date_input, str):
            logging.info(f"Date was given as a string: {date_input}")
            return DateTimeHandler.convert_string_to_standard_date(date_input)
        else:
            logging.error(f"Invalid date format: {date_input}")
            raise ValueError(f"Invalid date format: {date_input}")


class NMDBConfig:
    """
    Configuration class for NMDB data retrieval and processing.

    This class encapsulates configuration settings required for NMDB
    data retrieval, including date ranges for data collection, station
    identification, cache management, and data resolution settings. It
    ensures that all components of the NMDB data collection module use
    consistent and standardized configuration parameters.

    Parameters
    ----------
    start_date_wanted : str
        Start date for data retrieval, formatted as 'YYYY-MM-DD'.
    end_date_wanted : str
        End date for data retrieval, formatted as 'YYYY-MM-DD'.
    station : str, optional
        NMDB station code from which data is retrieved. Defaults to
        'JUNG'.
    cache_dir : str or None, optional
        Path to the cache directory for storing retrieved data. If None,
        uses the default OS cache directory.
    nmdb_table : str, optional
        Specific NMDB table to query data from. Defaults to 'revori'.
    resolution : str, optional
        Resolution of the data in minutes. Defaults to '60'.
    cache_exists : bool, optional
        Indicates whether cached data exists for the given parameters.
        Defaults to False.
    cache_start_date : str or None, optional
        Start date of the available cached data, formatted as
        'YYYY-MM-DD'. None if no cache exists.
    cache_end_date : str or None, optional
        End date of the available cached data, formatted as
        'YYYY-MM-DD'. None if no cache exists.
    start_date_needed : str or None, optional
        Start date for which data needs to be fetched, considering
        cached data. None if all data is cached.
    end_date_needed : str or None, optional
        End date for which data needs to be fetched, considering cached
        data. None if all data is cached.

    """

    @log_key_step("station", "nmdb_table", "resolution")
    def __init__(
        self,
        start_date_wanted,
        end_date_wanted,
        station="JUNG",
        reference_value=None,
        cache_dir=None,
        nmdb_table="revori",
        resolution="60",
        cache_exists=False,
        cache_start_date=None,
        cache_end_date=None,
        start_date_needed=None,
        end_date_needed=None,
    ):
        self._start_date_wanted = start_date_wanted
        self._end_date_wanted = end_date_wanted
        self._cache_dir = cache_dir
        self.station = station if station is not None else "JUNG"
        self.reference_value = reference_value
        self.nmdb_table = nmdb_table if nmdb_table is not None else "revori"
        self.resolution = resolution if resolution is not None else "60"
        self.cache_exists = cache_exists
        self.cache_start_date = cache_start_date
        self.cache_end_date = cache_end_date
        self.start_date_needed = start_date_needed
        self.end_date_needed = end_date_needed

    @property
    def start_date_wanted(self):
        return DateTimeHandler.standardize_date_input(self._start_date_wanted)

    @start_date_wanted.setter
    def start_date_wanted(self, value):
        self._start_date_wanted = DateTimeHandler.standardize_date_input(value)

    @property
    def end_date_wanted(self):
        return DateTimeHandler.standardize_date_input(self._end_date_wanted)

    @end_date_wanted.setter
    def end_date_wanted(self, value):
        self._end_date_wanted = DateTimeHandler.standardize_date_input(value)

    @property
    def cache_dir(self):
        GlobalConfig.create_cache_dir()
        if self._cache_dir is not None:
            return self._cache_dir
        else:
            return GlobalConfig.get_cache_dir()


class CacheHandler:
    """
    Class to handle cache management using downloaded NMDB data

    The cache handler managed file paths, naming of files, storage and
    deletion of the cache. As default it will be stored in the usual
    operating system cache location.

    Parameters
    ----------
    config : NMDBConfig
        An instance of the NMDBConfig class containing configuration
        settings

    Attributes
    ----------
    config : NMDBConfig
        Stores the configuration settings for NMDB data retrieval.
    _cache_file_path : Path or None
        The file path to the cache file, dynamically determined based on
        the NMDBConfig settings.

    Methods
    -------
    update_cache_file_path()
        Updates the cache file path based on the current NMDBConfig
        settings.
    check_cache_file_exists()
        Checks for the existence of the cache file and updates the
        configuration accordingly.
    read_cache()
        Reads the cached NMDB data from the file and returns it as a
        DataFrame.
    write_cache(cache_df)
        Writes a DataFrame to the cache file location.
    delete_cache()
        Deletes the cache file associated with the current NMDBConfig
        settings.
    check_cache_range()
        Determines the range of dates available in the cache and updates
        the NMDBConfig settings.

    Examples
    --------
    >>> config = NMDBConfig(start_date_wanted='2023-01-01',
    >>>             end_date_wanted='2023-01-31', station='JUNG')
    >>> cache_handler = CacheHandler(config)
    >>> cache_handler.update_cache_file_path()
    >>> print(cache_handler.cache_file_path)
    """

    def __init__(self, config):
        self.config = config
        self._cache_file_path = None
        self.update_cache_file_path()

    def update_cache_file_path(self):
        """Update the cache file path based on the current configuration."""
        self._cache_file_path = (
            Path(self.config.cache_dir)
            / f"nmdb_{self.config.station}_resolution_{self.config.resolution}"
            f"_nmdbtable_{self.config.nmdb_table}.csv"
        )

    @property
    def cache_file_path(self):
        return self._cache_file_path

    @cache_file_path.setter
    def cache_file_path(self, value):
        self._cache_file_path = value

    def check_cache_file_exists(self):
        """
        Checks the existence of the cache file and sets the property in
        config

        Returns
        -------
        None
        """
        if self.cache_file_path.exists():
            self.config.cache_exists = True

    def read_cache(self):
        """
        Reads cache nmdb file and formats index

        Returns
        -------
        df : pd.DataFrame
            DataFrame from the cache file
        """
        if self.config.cache_exists:
            df = pd.read_csv(self.cache_file_path)
            df["datetime"] = pd.to_datetime(df["datetime"])
            df.set_index("datetime", inplace=True)
            if df.index.tzinfo is None:
                df.index = df.index.tz_localize("UTC")
            else:
                df.index = df.index.tz_convert("UTC")
            return df

    def write_cache(self, cache_df):
        """
        Write NMDB data to the cache location using the cache_file_path
        attribute as a name.

        Parameters
        ----------
        cache_df : pd.DataFrame

        Returns
        -------
        None
        """
        if cache_df.empty:
            logging.warning("Attempting to write an empty DataFrame to cache.")
            return
        cache_df.to_csv(self.cache_file_path)

    def delete_cache(self):
        """
        Delete the cache file related to the current instance. E.g. if
        downloading hourly data for JUNG it will delete the file
        associated with hourly JUNG from the cache.

        Return
        ------
        None
        """
        if self.cache_file_path.exists():
            self.cache_file_path.unlink(missing_ok=True)
        logging.info("Cache file deleted")
        self.config.cache_exists = False

    def check_cache_range(self):
        """
        Function to find the range of data already available in the
        cache. It updates the config file depending on availability. It
        will either declare none existance of the cache, or update the
        start and end date of the cache.

        Returns
        -------
        None
        """
        self.check_cache_file_exists()
        if self.config.cache_exists:
            df_cache = self.read_cache()
            self.config.cache_start_date = df_cache.index.min().date()
            self.config.cache_end_date = df_cache.index.max().date()
        else:
            logging.info("There is no Cache file")
            self.config.cache_exists = False


class DataFetcher:
    """
    Class to handle sending external requests and fetching data.

    Parameters
    ----------
    config : NMDBConfig
        An instance of the configuration file.

    Methods
    -------
    get_ymd_from_date(date)
        static method which parses the date into seperate values to
        represent year, month and day
    create_nmdb_url()
        creates the url to request data from NMDB.eu, based on values in
        the configuration file
    fetch_data_http()
        uses the created url to request data from NMDB.eu and returns
        the text from the response
    parse_http_date(raw_data)
        uses the returned text from fetch_data_http() and parses it into
        a standard format. The format is a pd.Dataframe with a
        datetime.datetime index

    """

    def __init__(self, config):
        self.config = config

    @staticmethod
    def get_ymd_from_date(date):
        """
        Parses a given date into year, month, and day.

        Parameters
        ----------
        date : datetime or str
            The date to be parsed.

        Returns
        -------
        tuple
            Tuple containing the year, month, and day of the date.
        """
        standardized_date = str(
            DateTimeHandler.standardize_date_input(str(date))
        )
        year, month, day = standardized_date.split("-")
        return year, month, day

    def create_nmdb_url(self):
        """
        Creates the URL for obtaining the data using HTTP

        Returns
        -------
        url : str
            URL as a string
        """
        if self.config.start_date_needed is None:
            self.config.start_date_needed = self.config.start_date_wanted
        if self.config.end_date_needed is None:
            self.config.end_date_needed = self.config.end_date_wanted
        sy, sm, sd = self.get_ymd_from_date(self.config.start_date_needed)
        ey, em, ed = self.get_ymd_from_date(self.config.end_date_needed)

        nmdb_form = "wget"
        url = (
            f"http://nest.nmdb.eu/draw_graph.php?{nmdb_form}=1"
            f"&stations[]={self.config.station}"
            f"&tabchoice={self.config.nmdb_table}"
            f"&dtype=corr_for_efficiency"
            f"&tresolution={self.config.resolution}"
            f"&yunits=0&date_choice=bydate"
            f"&start_day={sd}&start_month={sm}&start_year={sy}"
            f"&start_hour=0&start_min=0&end_day={ed}&end_month={em}"
            f"&end_year={ey}&end_hour=23&end_min=59&output=ascii"
        )
        return url

    def fetch_data_http(self):
        """
        Fetches the data using http from NMDB.eu and processes it

        Returns
        -------
        Text : str
            Returns the text from the http site
        """
        url = self.create_nmdb_url()
        response = requests.get(url)
        response.raise_for_status()

        return response.text

    def parse_http_data(self, raw_data):
        """
        Parse the HTTP response data into a dataframe

        Parameters
        ----------
        raw_data : str
            The raw text file collected from NMDB.eu

        Returns
        -------
        pd.DataFrame
            A DataFrame with index DateTime and Counts per second

        Raises
        ------
        ValueError
            Raised if the requested date is not available at the
            specified NMDB station, indicated by a specific error
            message in the raw data.
        requests.exceptions.RequestException
            Raised if there's an issue parsing the HTTP response into a
            DataFrame, such as an incorrect format or network-related
            errors during the fetch.

        Examples
        --------
        Assuming an instance `nmdb_data_handler` of a class that
        includes this method:

        >>> df = nmdb_data_handler.fetch_and_parse_http_data()
        >>> print(df.head())

        """
        # if date has not been covered we raise an error
        if str(raw_data)[4:9] == "Sorry":
            raise ValueError(
                "Request date is not avalaible at ",
                self.config.station,
                " station, try other Neutron Monitoring station",
            )
        data = StringIO(raw_data)
        try:
            data = pd.read_csv(data, delimiter=";", comment="#")
            data.columns = ["count"]
            data.index.name = "datetime"
            data.index = pd.to_datetime(data.index).tz_localize("UTC")
            data.index = pd.to_datetime(data.index)
        except requests.exceptions.RequestException as e:
            logging.error(f"HTTP Request failed: {e}")
        except ValueError as e:
            logging.error(e)
        return data

    def fetch_and_parse_http_data(self):
        """
        Fetches raw NMDB data via HTTP and parses it into a pandas
        DataFrame.

        This method combines the functionalities of fetching NMDB data
        from the designated HTTP source and subsequently parsing that
        raw data into a structured DataFrame. It leverages
        `fetch_data_http` to retrieve the data and `parse_http_data` to
        transform it into a usable format.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the NMDB data, indexed by datetime
            with counts per second.

        Raises
        ------
        Refer to documentation for fetch_data_http() and
        parse_http_data()

        Examples
        --------
        Assuming an instance `nmdb_data_handler` of a class that
        includes this method:

        >>> df = nmdb_data_handler.fetch_and_parse_http_data()
        >>> print(df.head())
        """
        raw_data = self.fetch_data_http()
        return self.parse_http_data(raw_data)


class DataManager:
    """
    Manages the integration of cached and newly fetched NMDB data.

    This class is responsible for determining the necessity of fetching
    new NMDB data based on the existing cache and the desired date
    range. It also handles the combination of cached data with newly
    fetched data to provide a complete dataset for analysis.

    Parameters
    ----------
    config : NMDBConfig
        Configuration settings for NMDB data retrieval.
    cache_handler : CacheHandler
        Handles operations related to caching of NMDB data.
    data_fetcher : DataFetcher
        Responsible for fetching new data from NMDB.

    Attributes
    ----------
    need_data_before_cache : bool or None
        Indicates if data before the cached range is needed.
    need_data_after_cache : bool or None
        Indicates if data after the cached range is needed.

    Methods
    -------
    check_if_need_extra_data()
        Evaluates the need for fetching data outside the current cache
        range.
    set_dates_for_nmdb_download()
        Updates the configuration with the date ranges that need to be
        fetched.
    combine_cache_and_new_data(df_cache, df_download)
        Merges newly fetched data with existing cached data, ensuring no
        duplication.

    """

    def __init__(self, config, cache_handler, data_fetcher):
        """
        Initializes the DataManager with the given configuration, cache
        handler, and data fetcher.

        Parameters
        ----------
        config : NMDBConfig
            Configuration settings for NMDB data retrieval.
        cache_handler : CacheHandler
            Handles operations related to caching of NMDB data.
        data_fetcher : DataFetcher
            Responsible for fetching new data from NMDB.
        """

        self.config = config
        self.cache_handler = cache_handler
        self.data_fetcher = data_fetcher
        self.need_data_before_cache = None
        self.need_data_after_cache = None

    def check_if_need_extra_data(self):
        """
        Updates configuration instance with boolean values stating
        whether a download of data is required before or after the
        desired dates.

        Returns
        -------
        None
        """

        self.cache_handler.check_cache_range()
        start_date_wanted = pd.to_datetime(
            self.config.start_date_wanted
        ).date()
        end_date_wanted = pd.to_datetime(self.config.end_date_wanted).date()

        self.need_data_before_cache = (
            start_date_wanted < self.config.cache_start_date
        )
        self.need_data_after_cache = (
            end_date_wanted > self.config.cache_end_date
        )

    def set_dates_for_nmdb_download(self):
        """
        Updates the configuration instance with the download range for
        NMDB data based upon the desired data and the available data in
        the cache.

        Returns
        -------
        None
        """
        if self.need_data_before_cache and self.need_data_after_cache:
            self.config.start_date_needed = self.config.start_date_wanted
            self.config.end_date_needed = self.config.end_date_wanted

        elif self.need_data_before_cache:
            self.config.start_date_needed = self.config.start_date_wanted
            self.config.end_date_needed = self.config.cache_start_date

        elif self.need_data_after_cache:
            self.config.start_date_needed = self.config.cache_end_date
            self.config.end_date_needed = self.config.end_date_wanted

    def combine_cache_and_new_data(self, df_cache, df_download):
        """
        Combines cached and newly downloaded NMDB data into a single
        DataFrame, ensuring data continuity and no duplication.

        Parameters
        ----------
        df_cache : pd.DataFrame
            The DataFrame containing cached data.
        df_download : pd.DataFrame
            The DataFrame containing newly downloaded data.

        Returns
        -------
        pd.DataFrame
            The combined DataFrame, sorted by datetime.
        """
        if "datetime" not in df_cache.index.names:
            df_cache.set_index("datetime", inplace=True)
        if "datetime" not in df_download.index.names:
            df_download.index = df_download.index.tz_localize("UTC")

        if df_cache.index.tz is None:
            df_cache.index = df_cache.index.tz_localize("UTC")
        if df_download.index.tz is None:
            df_download.index = df_download.index.tz_localize("UTC")

        combined_df = pd.concat([df_cache, df_download])
        combined_df.reset_index(inplace=True)
        combined_df.drop_duplicates(
            subset="datetime", keep="first", inplace=True
        )
        combined_df.set_index("datetime", inplace=True)
        combined_df_sorted = combined_df.sort_index()
        return combined_df_sorted


class NMDBDataHandler:
    """
    Handles the retrieval and management of NMDB data.

    This class integrates the `CacheHandler`, `DataFetcher`, and
    `DataManager` to manage NMDB data. It ensures that data is fetched
    from the NMDB source only when necessary, preferring cached data to
    minimize network requests. The class handles cases where new data
    needs to be fetched either because it's not present in the cache or
    only partial data is available.

    Parameters
    ----------
    config : NMDBConfig
        Configuration settings for NMDB data retrieval, including
        desired date ranges, station information, and caching
        preferences.

    Attributes
    ----------
    config : NMDBConfig
        Stores the provided NMDB configuration settings.
    cache_handler : CacheHandler
        Manages caching operations for NMDB data.
    data_fetcher : DataFetcher
        Responsible for fetching new NMDB data when required.
    data_manager : DataManager
        Determines the need for and manages the retrieval of new data
        based on cache status and configuration settings.

    Methods
    -------
    collect_nmdb_data()
        Retrieves NMDB data, prioritizing cached data and fetching new
        data as needed. Returns a DataFrame containing the relevant NMDB
        data.

    Examples
    --------
    >>> config = NMDBConfig(start_date_wanted='2023-01-01',
    >>>             end_date_wanted='2023-01-31', station='JUNG')
    >>> nmdb_handler = NMDBDataHandler(config)
    >>> nmdb_data = nmdb_handler.collect_nmdb_data()
    >>> print(nmdb_data.head())
    """

    def __init__(self, config):
        """
        Initializes the NMDBDataHandler with the given NMDBConfig instance.

        Parameters
        ----------
        config : NMDBConfig
            Configuration settings for NMDB data retrieval.
        """
        self.config = config
        self.cache_handler = CacheHandler(config)
        self.data_fetcher = DataFetcher(config)
        self.data_manager = DataManager(
            config, self.cache_handler, self.data_fetcher
        )

    def collect_nmdb_data(self):
        """
        Collects NMDB data based on the specified configuration, using
        cached data when available and fetching new data as necessary.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing NMDB data for the requested range.
            This may be a combination of cached and newly fetched data,
            or solely from one source, depending on availability.

        Examples
        --------
        Assuming `config` has been defined and passed to
        `NMDBDataHandler`:

        >>> nmdb_data = nmdb_handler.collect_nmdb_data()
        >>> print(nmdb_data.head())

        Note: This example assumes that `nmdb_handler` has been
        instantiated with a valid `NMDBConfig`.
        """

        self.cache_handler.check_cache_file_exists()
        if self.config.cache_exists:
            self.cache_handler.check_cache_range()
            self.data_manager.check_if_need_extra_data()
            if (
                self.data_manager.need_data_before_cache is False
                and self.data_manager.need_data_after_cache is False
            ):
                core_logger.info("All data is present in the cache.")
                df_cache = self.cache_handler.read_cache()
                return df_cache

            else:
                self.data_manager.set_dates_for_nmdb_download()
                df_cache = self.cache_handler.read_cache()
                df_download = self.data_fetcher.fetch_and_parse_http_data()
                df_combined = self.data_manager.combine_cache_and_new_data(
                    df_cache, df_download
                )
                self.cache_handler.write_cache(df_combined)
                return df_combined
        else:
            core_logger.info(
                f"No cache file found at"
                f" {self.cache_handler.cache_file_path}."
            )
            df_download = self.data_fetcher.fetch_and_parse_http_data()
            self.cache_handler.write_cache(df_download)
            self.config.cache_exists = True
            self.cache_handler.cache_file = df_download
            return df_download



import streamlit as st
import pandas as pd

st.title(":material/info: Site information")

if not st.session_state["yaml_checked"]:
    st.warning("You need to select a configuration first.")
else:

    #######################
    st.subheader("1. Site")
    #######################

    c1, c2 = st.columns(2)

    # Name
    c1.text_input(
        label="Name",
        value=st.session_state["yaml"].sensor_config.sensor_info.name,
        key="input_sensor_name",
    )

    # Country
    c2.text_input(
        label="Country",
        value=st.session_state["yaml"].sensor_config.sensor_info.country,
        key="input_sensor_country",
    )

    # Identifier
    c1.text_input(
        label="Identifier",
        value=st.session_state["yaml"].sensor_config.sensor_info.identifier,
        key="input_sensor_identifier",
    )

    c1, c2 = st.columns(2)

    # Install date
    c1.date_input(
        label="Install date",
        value=st.session_state["yaml"].sensor_config.sensor_info.install_date,
        key="input_sensor_install_date",
    )

    # Timezone
    c2.number_input(
        label="Timezone",
        value=int(
            st.session_state["yaml"].sensor_config.sensor_info.time_zone
        ),
        key="input_sensor_time_zone",
        min_value=-12,
        max_value=14,
    )

    @st.fragment
    def make_location():
        c1, c2 = st.columns(2)

        # Latitude
        c1.number_input(
            label="Latitude",
            value=float(
                st.session_state["yaml"].sensor_config.sensor_info.latitude
            ),
            key="input_sensor_latitude",
            min_value=-180.0,
            max_value=180.0,
            format="%0.5f",
        )

        # Longitude
        c1.number_input(
            label="Longitude",
            value=float(
                st.session_state["yaml"].sensor_config.sensor_info.longitude
            ),
            key="input_sensor_longitude",
            min_value=-90.0,
            max_value=90.0,
            format="%0.5f",
        )

        # Elevation
        c1.number_input(
            label="Elevation",
            value=float(
                st.session_state["yaml"].sensor_config.sensor_info.elevation
            ),
            key="input_sensor_elevation",
            min_value=-1000.0,
            max_value=10000.0,
            format="%0.1f",
            step=1.0,
        )

        map_data = pd.DataFrame(
            dict(
                lat=[
                    st.session_state["input_sensor_latitude"],
                    st.session_state["input_sensor_latitude"],
                    st.session_state["input_sensor_latitude"],
                    # st.session_state["yaml"].sensor_config.sensor_info.latitude,
                ],
                lon=[
                    st.session_state["input_sensor_longitude"],
                    st.session_state["input_sensor_longitude"],
                    st.session_state["input_sensor_longitude"],
                    # st.session_state["yaml"].sensor_config.sensor_info.longitude,
                ],
                size=[5, 100, 200],
            )
        )
        c2.map(map_data, size="size", height=300)
        c2.write("Circle radius: 5, 100, 200 m")

    make_location()

    ######################################
    st.subheader("2. Physical parameters")
    ######################################

    c1, c2 = st.columns(2)

    # Avg Lattice Water
    c1.number_input(
        label="Avg Lattice Water",
        value=st.session_state[
            "yaml"
        ].sensor_config.sensor_info.avg_lattice_water,
        key="input_sensor_avg_lattice_water",
        min_value=0.0,
        max_value=0.5,
        step=0.001,
    )
    # Avg soil organic carbon
    c2.number_input(
        label="Avg Soil Organic Carbon",
        value=st.session_state[
            "yaml"
        ].sensor_config.sensor_info.avg_soil_organic_carbon,
        key="input_sensor_avg_soil_organic_carbon",
        min_value=0.0,
        max_value=0.5,
        step=0.001,
    )
    # Avg dry_soil_bulk_density
    c1.number_input(
        label="Avg dry_soil_bulk_density",
        value=st.session_state[
            "yaml"
        ].sensor_config.sensor_info.avg_dry_soil_bulk_density,
        key="input_sensor_avg_dry_soil_bulk_density",
        min_value=0.1,
        max_value=3.0,
        step=0.01,
    )
    # mean_pressure
    c2.number_input(
        label="Mean pressure",
        value=st.session_state["yaml"].sensor_config.sensor_info.mean_pressure,
        key="input_sensor_mean_pressure",
        min_value=1.0,
        max_value=1300.0,
        step=0.1,
    )
    # beta_coefficient
    c1.number_input(
        label="Beta coefficient",
        value=st.session_state[
            "yaml"
        ].sensor_config.sensor_info.beta_coefficient,
        key="input_sensor_beta_coefficient",
        min_value=0.0,
        max_value=1.0,
        step=0.0001,
    )
    # Site_cutoff_rigidity
    c2.number_input(
        label="Site cutoff rigidity",
        value=st.session_state[
            "yaml"
        ].sensor_config.sensor_info.site_cutoff_rigidity,
        key="input_sensor_site_cutoff_rigidity",
        min_value=0.0,
        max_value=19.0,
        step=0.01,
    )
    # N0
    c1.number_input(
        label="$N_0$",
        value=st.session_state["yaml"].sensor_config.sensor_info.N0,
        key="input_sensor_N0",
        min_value=1,
        max_value=100000,
        step=1,
    )
    ################################
    st.subheader("3. Apply changes")
    ################################

    from datetime import datetime

    if st.button("Apply"):
        st.session_state["yaml"].sensor_config.sensor_info.name = (
            st.session_state["input_sensor_name"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.country = (
            st.session_state["input_sensor_country"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.identifier = (
            st.session_state["input_sensor_identifier"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.install_date = (
            datetime.combine(
                st.session_state["input_sensor_install_date"],
                datetime.min.time(),
            )
        )
        st.session_state["yaml"].sensor_config.sensor_info.time_zone = (
            st.session_state["input_sensor_time_zone"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.latitude = (
            st.session_state["input_sensor_latitude"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.longitude = (
            st.session_state["input_sensor_longitude"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.elevation = (
            st.session_state["input_sensor_elevation"]
        )
        st.session_state[
            "yaml"
        ].sensor_config.sensor_info.site_cutoff_rigidity = st.session_state[
            "input_sensor_site_cutoff_rigidity"
        ]
        st.session_state[
            "yaml"
        ].sensor_config.sensor_info.avg_lattice_water = st.session_state[
            "input_sensor_avg_lattice_water"
        ]
        st.session_state[
            "yaml"
        ].sensor_config.sensor_info.avg_soil_organic_carbon = st.session_state[
            "input_sensor_avg_soil_organic_carbon"
        ]
        st.session_state[
            "yaml"
        ].sensor_config.sensor_info.avg_dry_soil_bulk_density = st.session_state[
            "input_sensor_avg_dry_soil_bulk_density"
        ]
        st.session_state["yaml"].sensor_config.sensor_info.beta_coefficient = (
            st.session_state["input_sensor_beta_coefficient"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.mean_pressure = (
            st.session_state["input_sensor_mean_pressure"]
        )
        st.session_state["yaml"].sensor_config.sensor_info.N0 = (
            st.session_state["input_sensor_N0"]
        )

        st.success("Changes applied :smile:")



"""
Functions related neutron corrections for changes to incoming neutron
intensity.

TODO: finish off the McJannetDesilets2023 method

Functions in this module:

    incoming_intensity_zreda_2012
    cutoff_rigidity_adjustment_to_jung
    incoming_intensity_adjustment_rc_corrected
    McjannetDesilets2023.local_gravity
    McjannetDesilets2023.local_air_pressure
    McjannetDesilets2023.tau
"""

import numpy as np
from neptoon.logging import get_logger

core_logger = get_logger()


def incoming_intensity_correction(
    incoming_intensity,
    ref_incoming_intensity,
    rc_scaling,
):
    """
    Calculate the correction factor for neutron counts based on the
    difference in incoming neutron intensity between current conditions
    and a reference.

    This function is based on the methodology introduced in Zreda
    (2012).

    Parameters
    ----------
    incoming_intensity : float
        Current incoming neutron intensity, in counts per time unit.
    ref_incoming_intensity : float
        Reference incoming neutron intensity, in counts per time unit.
    rc_scaling : float
        A scaling factor to account for difference in cut off rigidity
        between sensor and reference monitor

    Returns
    -------
    c_factor: float
        Correction factor to be multiplied with neutron counts.
    """
    intensity_ratio = ref_incoming_intensity / incoming_intensity
    c_factor = rc_scaling * (intensity_ratio - 1) + 1
    return c_factor


def rc_correction_hawdon(
    site_cutoff_rigidity,
    ref_monitor_cutoff_rigidity,
):
    """
    Creates adjustment parameter required to adjust for cutoff
    rigidities between a location and jungfraujoch. As described in
    Hawdon et al., 2014.

    Parameters
    ----------
    site_cutoff_rigidity : float
        Cutoff rigidity of CRNS site given in Gigavolts (Gv)
    reference_monitor_cutoff_rigidity : float, optional
        Cutoff rigidity of reference monitor given in Gigavolts (Gv)

    Returns
    -------
    rc_correction: float
        adjustment for differences in cutoff rigidity
    """

    rc_correction = (
        -0.075 * (site_cutoff_rigidity - ref_monitor_cutoff_rigidity) + 1
    )
    return rc_correction


class McjannetDesilets2023:
    """
    Provides methods for correcting incoming neutron intensity based on
    the research of McJannet and Desilets (2023).

    Reference:
        McJannet, D. L., & Desilets, D. (2023). Incoming Neutron Flux
        Corrections for Cosmic-Ray Soil and Snow Sensors Using the
        Global Neutron Monitor Network. Water Resources Research, 59,
        e2022WR033889. https://doi.org/10.1029/2022WR033889
    """

    @staticmethod
    def local_gravity(latitude, elevation):
        """
        Calculate the local gravitational acceleration based on latitude
        and elevation.

        Parameters
        ----------
        latitude : float
            Latitude in degrees
        elevation : float
            Elevation in meters (m)

        Returns
        -------
        gravity
            Local gravitational acceleration in meters per second
            squared (m/s^2).
        """
        rock_density = 2670
        g1 = 9.780327 * (
            1
            + 0.0053024 * np.sin(latitude / 180 * 3.141) ** 2
            - 0.0000058 * np.sin(2 * latitude / 180 * 3.141) ** 2
        )
        g2 = -3.086 * 10**-6 * elevation
        g3 = 4.194 * 10**-10 * rock_density * elevation
        gravity = g1 + g2 + g3

        return gravity

    @staticmethod
    def local_air_pressure(latitude, elevation):
        """
        Estimate the local atmospheric pressure based on latitude and elevation.

        Parameters
        ----------
        latitude : float
            Latitude in degrees
        elevation : float
            Elevation in meters (m)

        Returns
        -------
        p_ref: float
            Local atmospheric pressure in hectopascals (hPa).
        """
        g = McjannetDesilets2023.local_gravity(latitude, elevation)
        p_ref = 1013.25 * (1 + (-0.0065 / 288.15) * elevation) ** (
            (-g * 0.0289644) / (8.31432 * -0.0065)
        )
        p_ref = p_ref / 100

        return p_ref

    @staticmethod
    def tau(latitude, elevation, cut_off_rigidity):
        """
        Calculate the correction factor tau for neutron correction based
        on latitude, altitude, and cutoff rigidity.

        latitude : float
            Latitude in degrees
        elevation : float
            Elevation in meters (m)
        cut_off_rigidity: float
            Cutoff rigidity at the site in Gigavolts (Gv)

        Returns
        -------
        tau: float
            Tau correction factor to adjust difference between site and
            reference monitor cut off rigidity
        """
        c0 = -0.0009
        c1 = 1.7699
        c2 = 0.0064
        c3 = 1.8855
        c4 = 0.000013
        c5 = -1.2237
        eps = 1

        Rc_JUNG = 4.49
        atmospheric_depth_JUNG = 665.18
        tau_JUNG = (
            eps
            * 1
            * (c0 * atmospheric_depth_JUNG + c1)
            * (
                1
                - np.exp(
                    -(c2 * atmospheric_depth_JUNG + c3)
                    * Rc_JUNG ** (c4 * atmospheric_depth_JUNG + c5)
                )
            )
        )
        K_JUNG = 1 / tau_JUNG

        gravity = McjannetDesilets2023.local_gravity(latitude, elevation)
        pressure_ref = McjannetDesilets2023.local_air_pressure(
            latitude, elevation
        )
        pressure_ref = pressure_ref * 100  # convert to pascal
        amtmospheric_depth = 10 * pressure_ref / gravity
        tau = (
            eps
            * K_JUNG
            * (c0 * amtmospheric_depth + c1)
            * (
                1
                - np.exp(
                    -(c2 * amtmospheric_depth + c3)
                    * cut_off_rigidity ** (c4 * amtmospheric_depth + c5)
                )
            )
        )
        return tau



import logging
from functools import wraps
from inspect import signature
from pathlib import Path
from typing import Union
import yaml
import hashlib
import time

from neptoon.logging import get_logger


core_logger = get_logger()

"""
IDEAS:

The DataAuditLog is currently named by the user and a warning tells the
user when multiple logs of same name are shown (it would add logging to
old file). Perhaps in the final form it uses a universal name and gets
auto deleted when it is parsed into a YAML. That way the DataAuditLog
only there for the instance? For now I'll leave in the naming part but
perhaps do this.
"""


def log_key_step(*log_args):
    """
    Decorator which is used to record functions and values used in such
    functions.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                data_audit_log = DataAuditLog.get_instance()
            except Exception as e:
                core_logger.warning(
                    f"No DataAuditLog found: {e}/n"
                    "No DataAuditLog taking place"
                )
                data_audit_log = None

            if data_audit_log is not None:
                sig = signature(func)
                bound_arguments = sig.bind(*args, **kwargs)
                bound_arguments.apply_defaults()

                func_name = func.__name__
                if func_name == "__init__":
                    class_name = args[0].__class__.__name__
                    func_name = class_name

                data_audit_log_info = {
                    arg: bound_arguments.arguments[arg]
                    for arg in log_args
                    if arg in bound_arguments.arguments
                }

                data_audit_log.add_step(func_name, data_audit_log_info)

            return func(*args, **kwargs)

        return wrapper

    return decorator


class DataAuditLog:
    """
    The DataAuditLog in a singleton pattern
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        """
        Singleton pattern to ensure only one DataAuditLog is created at
        any one time. When instantiated it uses the init_data_audit_log
        method for creation.
        """

        if cls._instance is None:
            cls._instance = super(DataAuditLog, cls).__new__(cls)
        return cls._instance

    @classmethod
    def create(cls):
        """
        Creates a unique DataAuditLog instance with a specific filename.
        """
        if cls._instance is None:
            cls._instance = cls()
            cls._instance.init_data_audit_log()
        else:
            message = (
                "DataAuditLog instance already exists.\n "
                "Only one instance should be created for each\n"
                "processing run."
                "Use DataAuditLog.delete_instance() to remove it\n"
                "prior to running a new site."
            )
            core_logger.warning(message)
        return cls._instance

    @classmethod
    def get_instance(cls):
        return cls._instance

    @classmethod
    def delete_instance(cls):
        if cls._instance:
            cls._instance.close_log()
            cls._instance = None
        else:
            raise Exception("No instance exists for deletion")

    @classmethod
    def delete_log_file(cls):
        if cls._instance:
            cls._instance.file_delete()
        else:
            raise Exception("Could not find file for deletion")

    @classmethod
    def get_log_file_path(cls):
        if cls._instance and hasattr(cls._instance, "log_file_path"):
            return cls._instance.log_file_path
        else:
            core_logger.warning("No log_file_path available")

    @classmethod
    def create_log_folder(
        cls, site_name: str, custom_log_location: Union[Path, None] = None
    ):
        if cls._instance and hasattr(cls._instance, "log_file_path"):
            if site_name is None:
                raise Exception("You must select a name for the log.")
            timestamp = time.strftime("%Y-%m-%d %H-%M-%S")

            folder_name = f"{site_name} {timestamp}"
            if custom_log_location is None:
                log_location = Path.cwd()
            else:
                log_location = custom_log_location
            archive_folder = Path(
                log_location / "data_audit_log" / folder_name
            )
            archive_folder.mkdir(parents=True, exist_ok=True)
            cls._instance.archive_folder_location = archive_folder

    @classmethod
    def archive_data_audit(cls, site_name=None, custom_log_location=None):
        cls._instance.create_log_folder(
            site_name=site_name, custom_log_location=custom_log_location
        )

        archive_folder = cls._instance.archive_folder_location

        yaml_string = ParseDataAuditLog.parse_log_to_yaml_string(
            log_file_path=cls._instance.get_log_file_path()
        )
        yaml_hash = ParseDataAuditLog.hash_yaml_string(yaml_string=yaml_string)
        save_hash = archive_folder / "hash.txt"
        save_hash.write_text(yaml_hash)

        ParseDataAuditLog.save_yaml_string_to_file(
            yaml_string=yaml_string, yaml_save_location=archive_folder
        )

    @classmethod
    def archive_and_delete_log(cls, site_name=None, custom_log_location=None):
        cls._instance.archive_data_audit(
            site_name=site_name, custom_log_location=custom_log_location
        )
        cls._instance.delete_log_file()
        cls._instance.delete_instance()

    def init_data_audit_log(self):
        self.log_file_path = Path.cwd() / "DataAuditLog.log"
        self.logger = logging.getLogger("DataAuditLog")
        self.logger.propagate = False
        self.logger.setLevel(logging.INFO)
        data_audit_log_handler = logging.FileHandler(self.log_file_path)
        formatter = logging.Formatter("%(message)s")
        data_audit_log_handler.setFormatter(formatter)
        self.logger.addHandler(data_audit_log_handler)
        self.archive_folder_location = None

    def add_step(self, function_name, parameters):
        """
        Adds a record of a processing step to the log

        Parameters
        ----------
        function_name : str
            The name of the function being recorded.
        parameters : dict
            The parameters being applied in the function that are set to
            be recorded by the decorator: @key_steps_log()
        """
        params_yaml = parameters

        self.logger.info(
            f"function: {function_name} parameters: {params_yaml}"
        )

    def close_log(self):
        """
        Closes the log handlers attached to the logger.
        """
        for handler in self.logger.handlers:
            handler.close()
            self.logger.removeHandler(handler)

    def file_delete(self):
        """
        Deletes the log file associated with the DataAuditLog
        """
        try:
            self.log_file_path.unlink()
        except PermissionError:
            # TODO: should be a logger message/warning?
            print("Permission error: Log file is used by another process.")


class ParseDataAuditLog:
    """
    Parse the DataAuditLog file into a YAML
    """

    @staticmethod
    def parse_log_to_yaml_string(log_file_path):
        """
        Converts the DataAuditLog log file into a string in the YAML
        format.

        Parameters
        ----------
        log_file_path : pathlib.Path | str
            Path to the log file

        Returns
        -------
        yaml_str
            The log file, as a string, parsed into the YAML style.
        """
        functions_dict = {}
        with open(log_file_path, "r") as file:
            for line in file:
                func_part, params_str = line.split("parameters:")
                function_name = func_part.replace("function:", "").strip()
                params_str = params_str.strip()[1:-1]
                params_pairs = params_str.split(",")

                params_dict = {}
                for pair in params_pairs:
                    key, value = pair.split(":")
                    key = key.replace("'", "")
                    key = key.replace(" ", "")
                    value = value.replace("'", "")
                    value = value.replace(" ", "")
                    if value.replace(".", "", 1).isdigit():
                        value = float(value) if "." in value else int(value)
                    params_dict[key] = value

                if function_name not in functions_dict:
                    functions_dict[function_name] = []
                functions_dict[function_name].append(params_dict)

                # functions_dict[function_name] = params_dict
        yaml_str = yaml.dump(functions_dict)
        return yaml_str

    @staticmethod
    def save_yaml_string_to_file(yaml_string, yaml_save_location):
        with open((yaml_save_location / "audit.yaml"), "w") as yaml_file:
            yaml_file.write(yaml_string)

    @staticmethod
    def hash_yaml_string(yaml_string):
        """
        Creates a SHA256 hash based on the Yaml string provided.

        Parameters
        ----------
        yaml_str : str
            String in YAML style

        Returns
        -------
        SHA256 Hash
            The SHA256 Hash representing the string
        """
        return hashlib.sha256(yaml_string.encode("utf-8")).hexdigest()



import subprocess
from pathlib import Path
import typer
import platform
import sys
import shutil

app = typer.Typer()


def find_streamlit_executable():
    """Find streamlit executable in the current Python environment."""
    is_windows = platform.system() == "Windows"

    if is_windows:
        streamlit_path = (
            Path(sys.executable).parent / "Scripts" / "streamlit.exe"
        )
    else:
        streamlit_path = Path(sys.executable).parent / "streamlit"

    if streamlit_path.exists():
        return streamlit_path

    streamlit_in_path = shutil.which("streamlit")
    if streamlit_in_path:
        print(f"Found streamlit in PATH: {streamlit_in_path}")
        return Path(streamlit_in_path)

    print(f"Python executable: {sys.executable}")
    print(f"Looked in: {streamlit_path}")
    print("Searched PATH but couldn't find streamlit")

    return None


@app.command()
def main():
    """Launch the neptoon GUI application."""
    app_path = Path(__file__).parent.parent / "interface" / "gui.py"

    streamlit_path = find_streamlit_executable()

    if streamlit_path is None:
        raise ValueError(
            "Streamlit executable not found. "
            "Please ensure streamlit is installed in your environment."
        )

    subprocess.run(
        [
            str(streamlit_path),
            "run",
            str(app_path),
        ]
    )


if __name__ == "__main__":
    app(standalone_mode=False)



import pandas as pd
import math
from pathlib import Path
from typing import Union
import shutil
import json
import yaml
from typing import List
from magazine import Publish, Magazine
from neptoon.logging import get_logger
from neptoon.data_audit import DataAuditLog
from neptoon.config.configuration_input import (
    SensorInfo,
    SensorConfig,
    ProcessConfig,
)
from neptoon.utils.general_utils import validate_and_convert_file_path
from neptoon.visulisation.figures_handler import FigureHandler
from neptoon.columns import ColumnInfo

core_logger = get_logger()


class SaveAndArchiveOutputs:
    """
    Handles saving the outputs from neptoons in an organised way.

    Future Ideas:
    -------------
    - options to compress outputs (zip_output: bool = True)
    - cloud connection
    - bespoke output formats
    """

    def __init__(
        self,
        folder_name: str,
        processed_data_frame: pd.DataFrame,
        flag_data_frame: pd.DataFrame,
        sensor_info: SensorInfo,
        save_folder_location: str | Path | None = None,
        append_audit_log_hash_to_folder_name: bool = False,
        use_custom_column_names: bool = False,
        custom_column_names_dict: dict | None = None,
        append_timestamp: bool = True,
        figure_handler: FigureHandler | None = None,
        calib_df=None,
        magazine_active: bool = False,
    ):
        """
        Attributes

        Parameters
        ----------
        folder_name : str
            Desired name for the save folder
        processed_data_frame : pd.DataFrame
            The processed time series data
        flag_data_frame : pd.DataFrame
            The flag dataframe
        sensor_info : SensorInfo
            The SensorInfo object.
        save_folder_location : Union[str, Path], optional
            The folder where the data should be saved. If left as None
        append_audit_log_hash_to_folder_name : bool, optional
            The DataAuditLog gets converted to a hash, meaning sites
            processed the same way share a hash. This can be appended to
            the folder automatically helping to identify sites processed
            differently, by default False
        use_custom_column_names : bool, optional
             Whether to use custom column names, by default False
        custom_column_names_dict : dict, optional
            A dictionary to convert standard neptoon names into custom a
            custom naming convention, by default None
        append_timestamp: bool, optional, by default True
            Whether to append a timestamp to the folder name when
            saving.
        """
        self.folder_name = folder_name
        self.processed_data_frame = processed_data_frame
        self.flag_data_frame = flag_data_frame
        self.sensor_info = sensor_info
        self.save_folder_location = self._validate_save_folder(
            save_folder_location
        )
        self.append_audit_log_hash_to_folder_name = (
            append_audit_log_hash_to_folder_name
        )
        self.use_custom_column_names = use_custom_column_names
        self.custom_column_names_dict = custom_column_names_dict
        self.append_timestamp = append_timestamp
        self.full_folder_location = None
        self.figure_handler = figure_handler
        self.calib_df = calib_df
        self.magazine_active = magazine_active

    def _validate_save_folder(
        self,
        save_location: str | Path | None,
    ):
        """
        Converts string path to pathlib.Path. If given path is not an
        absolute path, saves data to the current working directory.

        Parameters
        ----------
        save_location : Union[str, Path]
            The location where the data should be saved. If a location
            other than the current working directory is desired, provide
            a full path (i.e., not a relative path).

        Returns
        -------
        pathlib.Path
            The pathlib.Path object
        """
        save_path = validate_and_convert_file_path(file_path=save_location)
        if save_path is None:
            save_path = validate_and_convert_file_path(file_path=Path.cwd())
        return save_path

    def create_save_folder(
        self,
    ):
        """
        Creates the folder location where the data will be saved.
        """

        # Make save folder if not already there
        try:
            self.save_folder_location.mkdir()
        except FileExistsError as e:
            message = f"Error: {e} \nFolder already exists."
            core_logger.info(message)

        if self.append_timestamp:
            from datetime import datetime

            # timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M")
            timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        else:
            timestamp = ""

        new_folder_name = self.folder_name + "_" + timestamp

        self.full_folder_location = self.save_folder_location / new_folder_name

        # Prevent overwriting station data
        try:
            self.full_folder_location.mkdir(parents=True)
        except FileExistsError as e:
            message = f"Error: {e} \nFolder already exists."
            core_logger.error(message)
            print(message + " Please change the folder name and try again.")
            raise FileExistsError

    def close_and_save_data_audit_log(
        self,
        append_hash: bool = False,
    ):
        """
        Handles closing the data audit log, producing the YAML output,
        and optionally appending a hash to the save location folder
        name.

        This function performs the following steps:
            1.  Archives and deletes the data audit log using
                DataAuditLog.archive_and_delete_log()

            2. If append_hash is True:
                a. Locates the hash.txt file in the data_audit_log
                   subfolder
                b. Reads the first 6 characters of the hash
                c. Renames the main folder to include this hash

        Parameters:
        -----------
        append_hash : bool, optional (default=False)
            If True, appends the first 6 characters of the hash from
            hash.txt to the folder name.

        """
        try:
            DataAuditLog.archive_and_delete_log(
                site_name=self.sensor_info.name,
                custom_log_location=self.full_folder_location,
            )
            if append_hash:
                new_folder_path = self.append_hash_to_folder_name(
                    self.full_folder_location
                )
                # update internal attribute
                self.full_folder_location = new_folder_path
        except AttributeError as e:
            message = f"{e}: DataAuditLog not present - skipping archive step"
            core_logger.info(message)
        except Exception as e:
            message = f"Unexpected error in DataAuditLog archiving: {e}"
            core_logger.error(message)

    def append_hash_to_folder_name(
        self,
        folder_path: Path | None,
    ):
        """
        Appends the first 6 characters of the hash from hash.txt to the
        folder name.

        Parameters:
        -----------
        folder_path : pathlib.Path
            The path to the folder to be renamed.

        Returns:
        --------
        pathlib.Path
            The path to the renamed folder.

        Raises:
        -------
        FileNotFoundError
            If the data audit log folder or hash.txt file is not found.
        PermissionError
            If permissions to access the folder are not available.
        """
        folder_name = folder_path.name
        data_audit_folder = folder_path / "data_audit_log"

        if not data_audit_folder.exists():
            raise FileNotFoundError(
                f"Data audit log folder not found: {data_audit_folder}"
            )

        try:
            unknown_folder_name = next(data_audit_folder.glob("*/"))
            hash_file = unknown_folder_name / "hash.txt"

            if not hash_file.exists():
                raise FileNotFoundError(f"Hash file not found: {hash_file}")

            with hash_file.open("r") as f:
                contents = f.read()

            hash_append = contents[:6]
            new_folder_name = f"{folder_name}_{hash_append}"
            new_folder_path = folder_path.parent / new_folder_name
            folder_path.rename(new_folder_path)

            return new_folder_path

        except StopIteration:
            raise FileNotFoundError(
                f"No subdirectories found in {data_audit_folder}"
            )
        except PermissionError:
            raise PermissionError(
                f"Permission denied when trying to access {folder_path}"
            )
        except Exception as e:
            raise RuntimeError(
                f"An error occurred while appending hash: {str(e)}"
            )

    def mask_bad_data(
        self,
    ):
        """
        Masks out flagged data with nan values
        """
        common_columns = self.flag_data_frame.columns.intersection(
            self.processed_data_frame.columns
        )
        if len(common_columns) < len(self.processed_data_frame.columns):
            core_logger.info(
                "processed_data_frame has additional columns that "
                "will not be masked."
            )
        mask = self.flag_data_frame == "UNFLAGGED"
        masked_df = self.processed_data_frame.copy()
        masked_df[~mask] = math.nan
        return masked_df

    def _save_figures(self):
        """
        Handles saving figures
        """

        figure_metadata = [
            fig_md for fig_md in self.figure_handler.temp_handler.get_figures()
        ]

        figure_folder = self.full_folder_location / "figures"
        figure_folder.mkdir(parents=True, exist_ok=True)
        for figure in figure_metadata:
            try:
                dest = figure_folder / f"{figure.name}.png"
                shutil.copy2(figure.path, dest)
            except FileNotFoundError as err:
                message = f"{figure.name} not found: {err}."
                core_logger.error(message)

    def _update_sensor_info(
        self,
        fields_to_check: List[str] = [
            "beta_coefficient",
            "mean_pressure",
        ],
        beta_col=str(ColumnInfo.Name.BETA_COEFFICIENT),
        mean_press_col=str(ColumnInfo.Name.MEAN_PRESSURE),
    ):
        """
        Updates SensorInfo if values where calulated during processing.

        Parameters
        ----------
        fields_to_check : List[str], optional
            A list of values in SensorInfo to check, by default [
            "beta_coefficient", "l_coefficient", "mean_pressure", ]
        beta_col : str, optional
            Beta Coefficient column name, by default
            str(ColumnInfo.Name.BETA_COEFFICIENT)
        mean_press_col : str, optional
            mean pressure column name, by default
            str(ColumnInfo.Name.MEAN_PRESSURE)
        """
        missing_fields = [
            field
            for field in fields_to_check
            if getattr(self.sensor_info, field) is None
        ]
        if (
            beta_col in self.processed_data_frame.columns
            and "beta_coefficient" in missing_fields
        ):
            beta_coeff = self.processed_data_frame[beta_col].iloc[0]
            self.sensor_info.beta_coefficient = round(beta_coeff, 4)
        if (
            mean_press_col in self.processed_data_frame.columns
            and "mean_pressure" in missing_fields
        ):
            mean_pressure = self.processed_data_frame[mean_press_col].iloc[0]
            self.sensor_info.mean_pressure = round(mean_pressure, 2)

    def _save_pdf(self, location: Path | str):
        """
        Exports the pdf built using magazine to the save folder.

        Parameters
        ----------
        location : Path
            The Path to the folder where the pdf is saved
        """
        if self.magazine_active:
            name = self.sensor_info.name
            save_location = location / f"Report-{name}.pdf"
            with Publish(str(save_location), f"{name} data") as pdf:
                pdf.add_topic("Neutron Correction")
                pdf.add_figure("Neutron Correction")
                pdf.add_topic("NMDB")
                pdf.add_figure("NMDB")
                pdf.add_topic("Soil Moisture")
                pdf.add_figure("Soil Moisture")
                pdf.add_topic("Atmospheric Conditions")
                pdf.add_figure("Atmospheric Conditions")
                pdf.add_topic("Calibration")
                pdf.add_figure("Calibration")
                pdf.add_topic("Data Preparation")
                pdf.add_figure("Data Preparation")

    def save_data_frames(self, file_name):
        """
        Saves various data frames as .csv files.
        """
        data_folder = self.full_folder_location / "data"
        data_folder.mkdir()
        self.processed_data_frame.to_csv(
            data_folder / f"{file_name}_processed_data.csv"
        )
        self.flag_data_frame.to_csv(data_folder / f"{file_name}_flags.csv")
        if self.calib_df is not None:
            self.calib_df.to_csv(data_folder / f"{file_name}_calibration.csv")

    def save_outputs(
        self,
        use_custom_column_names: bool = False,
    ):
        """
        The main function which chains the options.

        1. Create folder
        2. Mask time series
        3. Save time series
        4. Save flag df
        5. Optional: Save bespoke time series
        6. Optional: Save DAL
        7. Optional: Save Journalist
        8. Optional: rename folder
        9. Optional: compress data
        """
        if use_custom_column_names:
            if self.custom_column_names_dict is None:
                message = (
                    "Cannot use custom column names if no "
                    "column name dictionary supplied."
                )
                core_logger.error(message)
                print(message)
                raise ValueError
        file_name = self.sensor_info.name
        self.create_save_folder()
        self.save_data_frames(file_name=file_name)
        if self.figure_handler:
            self._save_figures()
        self._update_sensor_info()
        if Magazine.active:
            self._save_pdf(location=self.full_folder_location)
        self.close_and_save_data_audit_log(
            append_hash=self.append_audit_log_hash_to_folder_name
        )

    # ---- TODO below this line ----

    def save_custom_column_names(
        self,
    ):
        """
        WIP - save custom variable names using ColumnInfo.
        """
        pass


class ConfigSaver:
    """
    Saves the SensorConfig object as a yaml file.
    """

    def __init__(
        self,
        save_folder_location: Path | str,
        config: SensorConfig,
    ):
        self.save_folder_location = save_folder_location
        self.config = config

    def save(
        self,
    ):
        """
        Convert a Pydantic model to YAML file and save it.
        """
        if isinstance(self.config, SensorConfig):
            save_location = (
                self.save_folder_location / "updated_sensor_config.yaml"
            )
        if isinstance(self.config, ProcessConfig):
            save_location = (
                self.save_folder_location / "updated_process_config.yaml"
            )
        json_str = self.config.model_dump_json()
        data = json.loads(json_str)

        config_str = yaml.safe_dump(
            data,
            default_flow_style=False,
            allow_unicode=True,
            sort_keys=False,
            indent=2,
            default_style=None,
        )

        Path(save_location).write_text(config_str)



from neptoon.config.configuration_input import ConfigurationManager, BaseConfig

from pathlib import Path
import tempfile
import yaml
from copy import deepcopy
# import pytest


def test_returned_config_config_type():
    """
    Assert the config is stored as a BaseConfig type
    """
    mock_file_path = Path(__file__).parent / "mock_data" / "test_station.yaml"
    config_manager = ConfigurationManager()
    config_manager.load_configuration(mock_file_path)
    station_object = config_manager.get_config("sensor")
    assert isinstance(station_object, BaseConfig)


def test_configuration_management_integration_test():
    """
    The canary.

    Integration test to ensure the  total process is running as we
    expect. Will load, validate and get the test_station.yaml file.
    """
    mock_file_path = Path(__file__).parent / "mock_data" / "test_station.yaml"
    config_manager = ConfigurationManager()
    config_manager.load_configuration(mock_file_path)
    station_object = config_manager.get_config("sensor")

    assert station_object.sensor_info.country == "Germany"
    assert (
        station_object.time_series_data.key_column_info.thermal_neutron_columns
        is None
    )
    assert (
        station_object.calibration.key_column_names.profile_id == "Profile_ID"
    )

def test_working_dir_system():
    """
    Test the working_directory functionality using a
    platform-independent approach.
    
    This test verifies: 1. Default behavior: paths are resolved relative
    to config file location 2. Custom behavior: paths are resolved
    relative to specified working_directory
    """

    
    mock_file_path = Path(__file__).parent / "mock_data" / "test_station.yaml"
    
    with open(mock_file_path, 'r') as f:
        original_config = yaml.safe_load(f)
    
    # Create temporary directory for test
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir).resolve()
        
        # Create a copy of the YAML in the temp directory
        temp_yaml_path = temp_dir_path / "temp_test_station.yaml"
        
        # Make nested data directory
        data_dir_name = "custom_data_dir"
        data_dir = temp_dir_path / data_dir_name
        data_dir.mkdir()
        
        # 1. First test: Default behavior (no working_directory)
        with open(temp_yaml_path, 'w') as f:
            # Add a relative path to be resolved
            modified_config = deepcopy(original_config)
            if 'raw_data_parse_options' not in modified_config:
                modified_config['raw_data_parse_options'] = {}
            modified_config['raw_data_parse_options']['data_location'] = data_dir_name
            yaml.dump(modified_config, f)
        
        # Load the config
        config_manager = ConfigurationManager()
        config_manager.load_configuration(str(temp_yaml_path))
        sensor_config = config_manager.get_config("sensor")
        
        # Verify resolved path is a child of the config directory
        actual_path = Path(sensor_config.raw_data_parse_options.data_location).resolve()
        assert actual_path.is_relative_to(temp_dir_path), \
            f"Default resolution: Path {actual_path} should be under {temp_dir_path}"
        assert actual_path.name == data_dir_name, \
            f"Default resolution: Path should end with {data_dir_name}, got {actual_path.name}"
        
        # 2. Second test: With working_directory specified
        with tempfile.TemporaryDirectory() as working_dir:
            working_dir_path = Path(working_dir).resolve()
            
            # Create the same directory structure in working_dir
            working_data_dir = working_dir_path / data_dir_name
            working_data_dir.mkdir()
            
            # Modify config to include working_directory
            modified_config['working_directory'] = str(working_dir_path)
            
            with open(temp_yaml_path, 'w') as f:
                yaml.dump(modified_config, f)
            
            # Load the modified config
            config_manager = ConfigurationManager()
            config_manager.load_configuration(str(temp_yaml_path))
            sensor_config = config_manager.get_config("sensor")
            
            # Verify resolved path is a child of the working directory
            actual_path = Path(sensor_config.raw_data_parse_options.data_location).resolve()
            assert actual_path.is_relative_to(working_dir_path), \
                f"Working dir resolution: Path {actual_path} should be under {working_dir_path}"
            assert actual_path.name == data_dir_name, \
                f"Working dir resolution: Path should end with {data_dir_name}, got {actual_path.name}"


import pandas as pd
from typing import Literal, Optional
import pandera
import pandera.pandas as pa

###
from neptoon.columns import ColumnInfo
from neptoon.corrections import (
    neutrons_to_total_grav_soil_moisture_desilets_etal_2010,
    neutrons_to_total_grav_soil_moisture_koehli_etal_2021,
    Schroen2017,
)
from neptoon.data_prep.conversions import AbsoluteHumidityCreator
from neptoon.logging import get_logger
from neptoon.data_audit import log_key_step
from neptoon.quality_control.utils import _validate_df

core_logger = get_logger()

###
# Pandera schema for validation
###

base_input_schema = pa.DataFrameSchema(
    {
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT
        ): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT
        ): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
    },
    index=pa.Index(dtype=pd.DatetimeTZDtype(unit="ns", tz="UTC"), coerce=True),
    strict=False,
)

input_schema_koehli = pa.DataFrameSchema(
    {
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT
        ): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT
        ): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(ColumnInfo.Name.ABSOLUTE_HUMIDITY): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
    },
    index=pa.Index(dtype=pd.DatetimeTZDtype(unit="ns", tz="UTC"), coerce=True),
    strict=False,
)

output_schema = pa.DataFrameSchema(
    {
        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT
        ): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT
        ): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER): pa.Column(
            dtype=float,
            coerce=True,
            nullable=True,
        ),
        str(ColumnInfo.Name.SOIL_MOISTURE_MEASUREMENT_RADIUS): pa.Column(
            float,
            coerce=True,
            nullable=True,
        ),
        str(ColumnInfo.Name.SOIL_MOISTURE_MEASURMENT_DEPTH): pa.Column(
            float,
            coerce=True,
            nullable=True,
        ),
        str(ColumnInfo.Name.SOIL_MOISTURE_VOL): pa.Column(
            float,
            coerce=True,
            nullable=True,
        ),
        str(ColumnInfo.Name.SOIL_MOISTURE_GRAV): pa.Column(
            float,
            coerce=True,
            nullable=True,
        ),
    },
    index=pa.Index(dtype=pd.DatetimeTZDtype(unit="ns", tz="UTC"), coerce=True),
    strict=False,
)


class NeutronsToSM:
    """
    Class for converting a DataFrame containing corrected neutrons into
    soil moisture estimates. Includes calculations for depth.
    """

    def __init__(
        self,
        crns_data_frame: pd.DataFrame,
        n0: float,
        dry_soil_bulk_density: float = 1.4,
        lattice_water: float = 0,
        soil_organic_carbon: float = 0,
        corrected_neutrons_col_name: str = str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL
        ),
        soil_moisture_vol_col_name: str = str(
            ColumnInfo.Name.SOIL_MOISTURE_VOL
        ),
        soil_moisture_grav_col_name: str = str(
            ColumnInfo.Name.SOIL_MOISTURE_GRAV
        ),
        depth_column_name: str = str(
            ColumnInfo.Name.SOIL_MOISTURE_MEASURMENT_DEPTH
        ),
        radius_column_name: str = str(
            ColumnInfo.Name.SOIL_MOISTURE_MEASUREMENT_RADIUS
        ),
        conversion_theory: Literal[
            "desilets_etal_2010", "koehli_etal_2021"
        ] = "desilets_etal_2010",
        koehli_method_form: Literal[
            "Jan23_uranos",
            "Jan23_mcnpfull",
            "Mar12_atmprof",
            "Mar21_mcnp_drf",
            "Mar21_mcnp_ewin",
            "Mar21_uranos_drf",
            "Mar21_uranos_ewin",
            "Mar22_mcnp_drf_Jan",
            "Mar22_mcnp_ewin_gd",
            "Mar22_uranos_drf_gd",
            "Mar22_uranos_ewin_chi2",
            "Mar22_uranos_drf_h200m",
            "Aug08_mcnp_drf",
            "Aug08_mcnp_ewin",
            "Aug12_uranos_drf",
            "Aug12_uranos_ewin",
            "Aug13_uranos_atmprof",
            "Aug13_uranos_atmprof2",
        ] = "Mar21_uranos_drf",
        abs_air_humidity_col_name=str(ColumnInfo.Name.ABSOLUTE_HUMIDITY),
        air_pressure_col_name=str(ColumnInfo.Name.AIR_PRESSURE),
    ):
        """
        Attributes to be added to the class.

        Parameters
        ----------
        crns_data_frame : pd.DataFrame
            _description_
        n0 : float
            The n0 term
        dry_soil_bulk_density : float, optional
            in g/cm3, by default 1.4
        lattice_water : float, optional
            in decimal percent, by default 0
        soil_organic_carbon : float, optional
            in decimal percent, by default 0
        corrected_neutrons_col_name : str, optional
            column name where corrected neutrons are to be found, by
            default str(
            ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL )
        soil_moisture_vol_col_name : str, optional
            column name where volumetric soil moisture should be
            written, by default str(ColumnInfo.Name.SOIL_MOISTURE)
        depth_column_name : str, optional
            column name where depth estimates are written, by default
            str( ColumnInfo.Name.SOIL_MOISTURE_MEASURMENT_DEPTH )
        radius_column_name : str, optional
            column name where radius estimates are written, by default
            str( ColumnInfo.Name.SOIL_MOISTURE_MEASUREMENT_RADIUS )
        """
        self._crns_data_frame = _validate_df(
            crns_data_frame, schema=base_input_schema
        )
        self.n0 = n0
        # None vals can be passed from CRNSDataHub so leave these if else statements
        self.dry_soil_bulk_density = (
            dry_soil_bulk_density
            if dry_soil_bulk_density is not None
            else 1.42
        )
        self.lattice_water = lattice_water if lattice_water is not None else 0
        self.soil_organic_carbon = (
            soil_organic_carbon if soil_organic_carbon is not None else 0
        )
        self.water_equiv_soil_organic_carbon = self._convert_soc_to_wsom(
            self.soil_organic_carbon
        )
        self.corrected_neutrons_col_name = corrected_neutrons_col_name
        self.soil_moisture_grav_col_name = soil_moisture_grav_col_name
        self.soil_moisture_vol_col_name = soil_moisture_vol_col_name
        self.depth_column_name = depth_column_name
        self.radius_column_name = radius_column_name
        self.conversion_theory = conversion_theory
        self.koehli_method_form = koehli_method_form
        self.abs_air_humidity_col_name = abs_air_humidity_col_name
        self.air_pressure_col_name = air_pressure_col_name
        self.air_humidity_uncorrected = False

    @property
    def crns_data_frame(self):
        return self._crns_data_frame

    @crns_data_frame.setter
    def crns_data_frame(self, df):
        self._crns_data_frame = df

    @staticmethod
    def _convert_soc_to_wsom(soc):
        """
        Converts soil organic carbon values into water equivelant soil
        organic matter.

        doi: https://doi.org/10.1002/2013WR015138

        """
        return soc * 0.556

    def _check_if_humidity_correction_applied(
        self,
        auto_uncorrect=True,
    ):
        """
        Checks if corrected neutrons have already been corrected for
        humidity. If they have it, and auto_uncorrect is True, it will
        uncorrect them. The UTS function corrects for humidity in the
        conversion to soil moisture.

        Parameters
        ----------
        auto_uncorrect : bool, optional
            Whether to dis-apply the humidity correction, by default
            True
        """
        if (
            str(ColumnInfo.Name.HUMIDITY_CORRECTION)
            in self.crns_data_frame.columns
        ) and not self.air_humidity_uncorrected:
            if auto_uncorrect:
                self.crns_data_frame[
                    str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)
                ] = (
                    self.crns_data_frame[
                        str(ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_FINAL)
                    ]
                    / self.crns_data_frame[
                        str(ColumnInfo.Name.HUMIDITY_CORRECTION)
                    ]
                )
                self.air_humidity_uncorrected = True
                message = (
                    "Using köhli 2021 UTS method, but humidity correctio already applied.\n"
                    "Humidity correction was removed from corrected counts."
                )
                print(message)
                core_logger.info(message)

    def _ensure_abs_humidity_available(self):
        """
        Ensures that absolute himidity data is available for processing
        using Koehli et al., 2021.
        """
        abs_hum_creator = AbsoluteHumidityCreator(self.crns_data_frame)
        self.crns_data_frame = (
            abs_hum_creator.check_and_return_abs_hum_column()
        )

    def calculate_sm_estimates(
        self,
        neutron_data_column_name: str,
        soil_moisture_column_write_name_grav: Optional[str] = None,
        soil_moisture_column_write_name_vol: Optional[str] = None,
    ):
        """
        Calculates soil moisture estimates and adds them to the
        dataframe.

        This method applies the neutron-to-soil-moisture conversion for
        each row in the dataframe and stores the results in a new
        column.

        Parameters
        ----------
        neutron_data_column_name : str
            The name of the column containing neutron count data.
        soil_moisture_column_write_name_grav : str
            The name of the new column to store calculated gravimetric
            soil moisture values.
        soil_moisture_column_write_name_vol : str
            The name of the new column to store calculated volumetric
            soil moisture values.

        Returns
        -------
        None
            The method modifies the dataframe in-place.

        Notes
        -----
        This method assumes that the neutron data has been properly
        corrected and that all necessary parameters (n0, bulk density,
        etc.) have been set.
        """
        # Create a series of grav soil moisture (incl. LW and WESOC)
        if self.conversion_theory == "desilets_etal_2010":
            raw_grav = self.crns_data_frame.apply(
                lambda row: neutrons_to_total_grav_soil_moisture_desilets_etal_2010(
                    neutrons=row[neutron_data_column_name],
                    n0=self.n0,
                ),
                axis=1,
            )

        elif self.conversion_theory == "koehli_etal_2021":
            print(
                "Using Koehli et al., 2021 method for converting neutrons to soil moisture"
            )
            print("This takes a little longer... please stand by....")

            self._check_if_humidity_correction_applied(auto_uncorrect=True)
            self._ensure_abs_humidity_available()

            self.crns_data_frame = _validate_df(
                self.crns_data_frame, schema=input_schema_koehli
            )
            raw_grav = self.crns_data_frame.apply(
                lambda row: neutrons_to_total_grav_soil_moisture_koehli_etal_2021(
                    neutron_count=row[neutron_data_column_name],
                    n0=self.n0,
                    lattice_water=self.lattice_water,
                    abs_air_humidity=row[self.abs_air_humidity_col_name],
                    water_equiv_soil_organic_carbon=self.water_equiv_soil_organic_carbon,
                    koehli_method_form=self.koehli_method_form,
                ),
                axis=1,
            )

        grav_sm = (
            raw_grav
            - self.lattice_water
            - self.water_equiv_soil_organic_carbon
        )

        if soil_moisture_column_write_name_grav:
            self.crns_data_frame[soil_moisture_column_write_name_grav] = (
                grav_sm
            )

        if soil_moisture_column_write_name_vol:
            self.crns_data_frame[soil_moisture_column_write_name_vol] = (
                grav_sm * self.dry_soil_bulk_density
            )

    def _from_grav_to_vol_sm(self, grav_sm, bulk_density):
        """
        Converts gravimetric to volumetric soil moisture

        Parameters
        ----------
        grav_sm : float
            Gravimetric soil moisture
        bulk_density : float
            Bulk density

        Returns
        -------
        float
            Volumetric Soil moisture
        """

        return grav_sm * bulk_density

    def calculate_uncertainty_of_sm_estimates(self):
        """
        Produces uncertainty estimates of soil mositure.
        """
        self.calculate_sm_estimates(
            neutron_data_column_name=str(
                ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_LOWER_COUNT
            ),
            soil_moisture_column_write_name_vol=str(
                ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_UPPER
            ),
        )
        self.calculate_sm_estimates(
            neutron_data_column_name=str(
                ColumnInfo.Name.CORRECTED_EPI_NEUTRON_COUNT_UPPER_COUNT
            ),
            soil_moisture_column_write_name_vol=str(
                ColumnInfo.Name.SOIL_MOISTURE_UNCERTAINTY_VOL_LOWER
            ),
        )

    @log_key_step("radius")
    def calculate_depth_of_measurement(
        self,
        radius: float = 50,
    ):
        """
        Creates a column with the calculated depth of measurement (cm)

        Parameters
        ----------
        radius : float, optional
            The default radius of measurement (avg), by default 50
        """
        self.crns_data_frame[self.depth_column_name] = (
            self.crns_data_frame.apply(
                lambda row: Schroen2017.calculate_measurement_depth(
                    distance=radius,
                    bulk_density=self.dry_soil_bulk_density,
                    volumetric_soil_moisture=row[
                        self.soil_moisture_vol_col_name
                    ],
                ),
                axis=1,
            )
        )

    def calculate_horizontal_footprint(self):
        """
        Creates a column with the calculated radius of measurement (m).
        """

        self.crns_data_frame[self.radius_column_name] = (
            self.crns_data_frame.apply(
                lambda row: Schroen2017.calculate_footprint_radius(
                    volumetric_soil_moisture=row[
                        self.soil_moisture_vol_col_name
                    ],
                    abs_air_humidity=row[self.abs_air_humidity_col_name],
                    atmospheric_pressure=row[self.air_pressure_col_name],
                ),
                axis=1,
            )
        )

    def calculate_all_soil_moisture_data(self):
        """
        TODO: Overall process method which will chain together the other
        methods to produce a fully developed DataFrame.
        """

        self.calculate_sm_estimates(
            neutron_data_column_name=self.corrected_neutrons_col_name,
            soil_moisture_column_write_name_grav=self.soil_moisture_grav_col_name,
            soil_moisture_column_write_name_vol=self.soil_moisture_vol_col_name,
        )
        self.calculate_uncertainty_of_sm_estimates()
        self.calculate_depth_of_measurement()
        self.calculate_horizontal_footprint()
        self.crns_data_frame = _validate_df(
            df=self.crns_data_frame, schema=output_schema
        )

    def return_data_frame(self):
        """
        Returns the crns DataFrame

        Returns
        -------
        pd.DataFrame
            The stored DataFrame
        """
        return self.crns_data_frame



from pathlib import Path
import pandas as pd
from typing import Literal

from neptoon.io.read.data_ingest import (
    FileCollectionConfig,
    ManageFileCollection,
    ParseFilesIntoDataFrame,
    InputDataFrameFormattingConfig,
    FormatDataForCRNSDataHub,
    validate_and_convert_file_path,
)
from neptoon.config.configuration_input import ConfigurationManager, BaseConfig
from neptoon.logging import get_logger

core_logger = get_logger()


def _get_config_section(
    configuration_object: ConfigurationManager,
    wanted_config: Literal["sensor", "process"],
):
    """
    Retrieves the specific config from the configuration manager.

    Parameters
    ----------
    configuration_object : ConfigurationManager
        The configuration object containing loaded configs
    wanted_object : Literal["sensor", "process"]
        The type of configuration object to retrieve

    Returns
    -------
    Optional[BaseConfig]
        The requested configuration object if found, None otherwise
    """
    try:
        return configuration_object.get_config(name=wanted_config)
    except (AttributeError, KeyError):
        core_logger.info(f"Configuration for {wanted_config} not found.")
        return None


def _return_config(
    path_to_config,
    config_to_return: Literal["sensor", "process"],
):
    """
    Loads the config file into a ConfigurationManager and returns the
    sensor config part

    Parameters
    ----------
    path_to_config : str | Path
        Path to config file (in yaml format)
    config_to_return: Literal["sensor", "process"]
        name of config to return
    """
    configuration_object = ConfigurationManager()
    configuration_object.load_configuration(
        file_path=path_to_config,
    )
    return _get_config_section(
        configuration_object=configuration_object,
        wanted_config=config_to_return,
    )


def _return_config(
    path_to_config,
    config_to_return: Literal["sensor", "process"],
):
    """
    Loads the config file into a ConfigurationManager and returns the
    sensor config part

    Parameters
    ----------
    path_to_config : str | Path
        Path to config file (in yaml format)
    config_to_return: Literal["sensor", "process"]
        name of config to return
    """
    configuration_object = ConfigurationManager()
    configuration_object.load_configuration(
        file_path=path_to_config,
    )
    return _get_config_section(
        configuration_object=configuration_object,
        wanted_config=config_to_return,
    )


class DataHubFromConfig:
    """
    Creates a DataHub instance using a configuration file.

    This class handles the configuration and initialization of a
    CRNSDataHub using a sensor configuration file. It manages raw data
    parsing, time series preparation, and final hub creation.

    Example:
    --------
    >>> # Method 1: Using a path to sensor configuration file
    >>> sensor_config_path = "/path/to/configurations/A101_station.yaml"
    >>> data_hub_creator = DataHubFromConfig(path_to_sensor_config=sensor_config_path)
    >>> data_hub = data_hub_creator.create_data_hub()
    >>>
    >>> # Method 2: Using a pre-configured ConfigurationManager
    >>> config_manager = ConfigurationManager()
    >>> config_manager.load_configuration(file_path=sensor_config_path)
    >>> data_hub_creator = DataHubFromConfig(configuration_object=config_manager)
    >>> data_hub = data_hub_creator.create_data_hub()
    >>>
    >>> # After creating the data hub, you can proceed with operations:
    >>> data_hub.attach_nmdb_data(station="JUNG")
    >>> data_hub.prepare_static_values()
    """

    def __init__(
        self,
        path_to_sensor_config: str | Path = None,
        configuration_object: ConfigurationManager = None,
        sensor_config: BaseConfig = None,  # Internal use
    ):
        """

        Parameters
        ----------
        path_to_sensor_config : str | Path, optional
            path where sensor config file is found , by default None
        configuration_object : ConfigurationManager, optional
            ConfigurationManager, presumed to contain a sensor config
            object, by default None
        sensor_config : BaseConfig, optional
            SensorConfig directly supplied (internal use with
            ProcessWithConfig), by default None
        """
        self.configuration_object = None
        self.sensor_config = None

        path_to_sensor_config = validate_and_convert_file_path(
            path_to_sensor_config
        )
        self.sensor_config = self._initialise_configuration(
            path_to_sensor_config=path_to_sensor_config,
            configuration_object=configuration_object,
            sensor_config=sensor_config,
        )

    def _initialise_configuration(
        self,
        path_to_sensor_config: str | Path,
        configuration_object: ConfigurationManager,
        sensor_config: BaseConfig,
    ):
        """
        Organises the initialisation steps to ensure a configuration
        object is available

        Parameters
        ----------
        path_to_sensor_config : str | Path, optional
            path where sensor config file is found , by default None
        configuration_object : ConfigurationManager, optional
            ConfigurationManager, presumed to contain a sensor config
            object, by default None
        sensor_config : BaseConfig, optional
            SensorConfig directly supplied (internal use with
            ProcessWithConfig), by default None

        Returns
        -------
        sensor_config
            sensor config file
        """
        if sensor_config is not None:
            return sensor_config
        elif configuration_object:
            sensor_config = _get_config_section(
                configuration_object=self.configuration_object,
                wanted_config="sensor",
            )
            return sensor_config
        elif path_to_sensor_config:
            sensor_config = _return_config(
                path_to_config=path_to_sensor_config, config_to_return="sensor"
            )
            return sensor_config
        else:
            self._no_data_given_error()

    def _no_data_given_error(self):
        """
        Raise ValueError if nothing supplied

        Raises
        ------
        ValueError
            No Data
        """
        message = (
            "Please provide either a path_to_sensor_config"
            " or a configuration_object"
        )
        core_logger.error(message)
        raise ValueError(message)

    def _parse_raw_data(
        self,
    ):
        """
        Parses raw data files.

        Returns
        -------
        pd.DataFrame
            DataFrame from raw files.
        """
        # create tmp object for more readable code
        tmp = self.sensor_config.raw_data_parse_options

        file_collection_config = FileCollectionConfig(
            data_location=tmp.data_location,
            column_names=tmp.column_names,
            prefix=tmp.prefix,
            suffix=tmp.suffix,
            encoding=tmp.encoding,
            skip_lines=tmp.skip_lines,
            separator=tmp.separator,
            decimal=tmp.decimal,
            skip_initial_space=tmp.skip_initial_space,
            parser_kw_strip_left=tmp.parser_kw.strip_left,
            parser_kw_digit_first=tmp.parser_kw.digit_first,
            starts_with=tmp.starts_with,
            multi_header=tmp.multi_header,
            strip_names=tmp.strip_names,
            remove_prefix=tmp.remove_prefix,
        )
        file_manager = ManageFileCollection(config=file_collection_config)
        file_manager.get_list_of_files()
        file_manager.filter_files()
        file_parser = ParseFilesIntoDataFrame(
            file_manager=file_manager, config=file_collection_config
        )
        parsed_data = file_parser.make_dataframe()

        return parsed_data

    def _import_data(
        self,
    ):
        """
        Imports data using information in the config file. If raw data
        requires parsing it will do this. If not, it is presumed the
        data is already available in a single csv file. It then uses the
        supplied information in the YAML files to prepare this for use
        in neptoon.

        Returns
        -------
        pd.DataFrame
            Prepared DataFrame
        """
        if self.sensor_config.raw_data_parse_options.parse_raw_data:
            raw_data_parsed = self._parse_raw_data()
        else:
            raw_data_parsed = pd.read_csv(
                validate_and_convert_file_path(
                    file_path=self.sensor_config.time_series_data.path_to_data,
                )
            )
        df = self._prepare_time_series(raw_data_parsed=raw_data_parsed)
        return df

    def _prepare_time_series(
        self,
        raw_data_parsed: pd.DataFrame,
    ):
        """
        Method for preparing the time series data.

        Returns
        -------
        pd.DataFrame
            Returns a formatted dataframe
        """
        input_formatter_config = InputDataFrameFormattingConfig()
        input_formatter_config.config_info = self.sensor_config
        input_formatter_config.build_from_config()

        data_formatter = FormatDataForCRNSDataHub(
            data_frame=raw_data_parsed,
            config=input_formatter_config,
        )
        df = data_formatter.format_data_and_return_data_frame()
        return df

    def create_data_hub(self):
        """
        Creates a CRNSDataHub using the supplied configuration
        information.

        This method processes raw data according to the configuration
        settings, formats it appropriately, and initializes a new
        CRNSDataHub instance.

        Returns
        -------
        CRNSDataHub
            A fully configured CRNSDataHub instance ready for further
            processing
        """
        # import here to avoid circular dependency
        from neptoon.hub import CRNSDataHub

        return CRNSDataHub(
            crns_data_frame=self._import_data(),
            sensor_info=self.sensor_config.sensor_info,
        )



import numpy as np
import pandas as pd
from typing import Literal


def neutrons_to_total_grav_soil_moisture_desilets_etal_2010(
    neutrons,
    n0=1000,
    a0=0.0808,
    a1=0.372,
    a2=0.115,
):
    """
    Converts neutrons to total gravimetric soil moisture following the
    Desilets et al., 2010 forumla. When we refer total gravimetric we
    mean that it's the amount including lattice water and water
    equivelant soil organic carbon.

    Parameters
    ----------
    neutrons : float
        corrected neutron count
    n0 : int | float, optional
        the n0 calibration term, by default 1000
    a0 : float, optional
        constant, by default 0.0808
    a1 : float, optional
        constant, by default 0.372
    a2 : float, optional
        constant, by default 0.115

    Returns
    -------
    float
        calculated gravimetric soil moisture value g/g
    """
    return a0 / (neutrons / n0 - a1) - a2


def neutrons_to_vol_soil_moisture_desilets_etal_2010(
    neutron_count: float,
    n0: float,
    dry_soil_bulk_density: float,
    lattice_water: float,
    water_equiv_soil_organic_carbon: float,
    a0: float = 0.0808,
    a1: float = 0.372,
    a2: float = 0.115,
):
    """
    Converts corrected neutrons counts into volumetric soil moisture
    following the Desilets et al., 2010 forumla.

    doi: http://dx.doi.org/10.1029/2009WR008726

    Parameters
    ----------
    neutron_count : int
        Neutron count in counts per hour (cph)
    n0 : int
        N0 calibration term
    bulk_density : float
        dry soil bulk density of the soil in grams per cubic centimer
        e.g. 1.4 (g/cm^3)
    lattice_water : float
        lattice water - decimal percent e.g. 0.002
    water_equiv_soil_organic_carbon : float
        water equivelant soil organic carbon - decimal percent e.g, 0.02
    a0 : float
        constant
    a1 : float
        constant
    a2 : float
        constant
    """
    return (
        ((a0) / ((neutron_count / n0) - a1))
        - (a2)
        - lattice_water
        - water_equiv_soil_organic_carbon
    ) * dry_soil_bulk_density


def reformulated_neutrons_to_grav_soil_moisture_desilets_2010(
    neutron_count: float,
    n0: float,
    lattice_water: float,
    water_equiv_soil_organic_carbon: float,
    a0: float = 0.0808,
    a1: float = 0.372,
    a2: float = 0.115,
):
    """
    Converts corrected neutrons counts into gravimetric soil moisture
    following the reforumlated version of the desilets equation outlined
    in Köhli et al. 2021

    https://doi.org/10.3389/frwa.2020.544847

    Parameters
    ----------
    a0 : float
        Constant
    a1 : float
        Constant
    a2 : float
        Constant
    neutron_count : int
        Neutron count in counts per hour (cph)
    n0 : int
        N0 number given as maximum number of neutrons possible over a 1
        hour integration.
    lattice_water : float
        Lattice water - decimal percent e.g. 0.002
    water_equiv_soil_organic_carbon : float
        Water equivelant soil organic carbon - decimal percent e.g, 0.02

    Returns
    -------
    volumetric_sm : float
        Volumetric soil moisture
    """
    nmax = n0 * ((a0 + (a1 * a2)) / (a2))
    ah0 = -a2
    ah1 = (a1 * a2) / (a0 + (a1 * a2))
    volumetric_sm = (
        (ah0 * ((1 - (neutron_count / nmax)) / (ah1 - (neutron_count / nmax))))
        - lattice_water
        - water_equiv_soil_organic_carbon
    )
    return volumetric_sm


def neutrons_to_total_grav_soil_moisture_koehli_etal_2021(
    neutron_count: float,
    n0: float,
    abs_air_humidity: float,
    lattice_water: float = 0.0,
    water_equiv_soil_organic_carbon: float = 0.0,
    koehli_method_form: Literal[
        "Jan23_uranos",
        "Jan23_mcnpfull",
        "Mar12_atmprof",
        "Mar21_mcnp_drf",
        "Mar21_mcnp_ewin",
        "Mar21_uranos_drf",
        "Mar21_uranos_ewin",
        "Mar22_mcnp_drf_Jan",
        "Mar22_mcnp_ewin_gd",
        "Mar22_uranos_drf_gd",
        "Mar22_uranos_ewin_chi2",
        "Mar22_uranos_drf_h200m",
        "Aug08_mcnp_drf",
        "Aug08_mcnp_ewin",
        "Aug12_uranos_drf",
        "Aug12_uranos_ewin",
        "Aug13_uranos_atmprof",
        "Aug13_uranos_atmprof2",
    ] = "Mar21_uranos_drf",
):
    """
    Converts corrected neutrons counts into volumetric soil moisture
    following the Universal Transport Solution (UTS) method outlined in
    Köhli et al. 2021

    https://doi.org/10.3389/frwa.2020.544847


    Parameters
    ----------
    neutron_count : float
        Neutron count in counts per hour (cph)
    n0 : float
        N0 calibration term
    abs_air_humidity : float
         absolute air humidity (g/cm3))
    lattice_water : float
        lattice water - decimal percent e.g. 0.002
    water_equiv_soil_organic_carbon : float
        water equivelant soil organic carbon - decimal percent e.g, 0.02

    Example
    -------
    convert_neutrons_to_soil_moisture_uts(
        neutron_count=2000, n0=3000, air_humidity=5
    )

    """
    if pd.isna(neutron_count) or pd.isna(abs_air_humidity):
        return np.nan

    gravimetric_soil_moisture_0 = 0.0
    gravimetric_soil_moisture_1 = 2.0
    while gravimetric_soil_moisture_1 - gravimetric_soil_moisture_0 > 0.0001:
        gravimetric_soil_moisture_2 = (0.5 * gravimetric_soil_moisture_0) + (
            0.5 * gravimetric_soil_moisture_1
        )
        n2 = gravimetric_soil_moisture_to_neutrons_koehli_etal_2021(
            gravimetric_sm=gravimetric_soil_moisture_2,
            abs_air_humidity=abs_air_humidity,
            n0=n0,
            koehli_method_form=koehli_method_form,
            offset=lattice_water + water_equiv_soil_organic_carbon,
        )
        if neutron_count < n2:
            gravimetric_soil_moisture_0 = gravimetric_soil_moisture_2
        else:
            gravimetric_soil_moisture_1 = gravimetric_soil_moisture_2
    return gravimetric_soil_moisture_2


def gravimetric_soil_moisture_to_neutrons_koehli_etal_2021(
    gravimetric_sm: float,
    abs_air_humidity: float,
    n0: float,
    offset: float = 0.0,
    koehli_method_form: Literal[
        "Jan23_uranos",
        "Jan23_mcnpfull",
        "Mar12_atmprof",
        "Mar21_mcnp_drf",
        "Mar21_mcnp_ewin",
        "Mar21_uranos_drf",
        "Mar21_uranos_ewin",
        "Mar22_mcnp_drf_Jan",
        "Mar22_mcnp_ewin_gd",
        "Mar22_uranos_drf_gd",
        "Mar22_uranos_ewin_chi2",
        "Mar22_uranos_drf_h200m",
        "Aug08_mcnp_drf",
        "Aug08_mcnp_ewin",
        "Aug12_uranos_drf",
        "Aug12_uranos_ewin",
        "Aug13_uranos_atmprof",
        "Aug13_uranos_atmprof2",
    ] = "Mar21_uranos_drf",
    biomass: float = 0.0,
):
    """
    Convert soil moisture to neutrons following following the method
    outlined in Köhli et al. 2021

    https://doi.org/10.3389/frwa.2020.544847

    Parameters
    ----------
    gravimetric_sm : float
        soil moisture gravimetric g/g
    abs_air_humidity : float
        absolute air humidity at the site (g/cm3)
    n0 : float
        n0 calibration term
    offset : float
        offset to apply to soil moisture. e.g., to account for lattice
        water
    bulk_density : float
        dry soil bulk density g/cm^3
    method : str
        The method to apply. See reference. default Mar21_uranos_drf
    biomass : float
        The biomass at the site kg/m^2

    """

    soil_moisture_total = gravimetric_sm + offset
    if soil_moisture_total == 0.0:
        soil_moisture_total = 0.001
    p = []

    if koehli_method_form == "Jan23_uranos":
        p = [
            4.2580,
            0.0212,
            0.206,
            1.776,
            0.241,
            -0.00058,
            -0.02800,
            0.0003200,
            -0.0000000180,
        ]
    elif koehli_method_form == "Jan23_mcnpfull":
        p = [
            7.0000,
            0.0250,
            0.233,
            4.325,
            0.156,
            -0.00066,
            -0.01200,
            0.0004100,
            -0.0000000410,
        ]
    elif koehli_method_form == "Mar12_atmprof":
        p = [
            4.4775,
            0.0230,
            0.217,
            1.540,
            0.213,
            -0.00022,
            -0.03800,
            0.0003100,
            -0.0000000003,
        ]
    elif koehli_method_form == "Mar21_mcnp_drf":
        p = [
            1.0940,
            0.0280,
            0.254,
            3.537,
            0.139,
            -0.00140,
            -0.00880,
            0.0001150,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar21_mcnp_ewin":
        p = [
            1.2650,
            0.0259,
            0.135,
            1.237,
            0.063,
            -0.00021,
            -0.01170,
            0.0001200,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar21_uranos_drf":
        p = [
            1.0240,
            0.0226,
            0.207,
            1.625,
            0.235,
            -0.00290,
            -0.00930,
            0.0000740,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar21_uranos_ewin":
        p = [
            1.2230,
            0.0185,
            0.142,
            2.568,
            0.155,
            -0.00047,
            -0.01190,
            0.0000920,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar22_mcnp_drf_Jan":
        p = [
            1.0820,
            0.0250,
            0.235,
            4.360,
            0.156,
            -0.00071,
            -0.00610,
            0.0000500,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar22_mcnp_ewin_gd":
        p = [
            1.1630,
            0.0244,
            0.182,
            4.358,
            0.118,
            -0.00046,
            -0.00747,
            0.0000580,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar22_uranos_drf_gd":
        p = [
            1.1180,
            0.0221,
            0.173,
            2.300,
            0.184,
            -0.00064,
            -0.01000,
            0.0000810,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar22_uranos_ewin_chi2":
        p = [
            1.0220,
            0.0218,
            0.199,
            1.647,
            0.243,
            -0.00029,
            -0.00960,
            0.0000780,
            0.0000000000,
        ]
    elif koehli_method_form == "Mar22_uranos_drf_h200m":
        p = [
            1.0210,
            0.0222,
            0.203,
            1.600,
            0.244,
            -0.00061,
            -0.00930,
            0.0000740,
            0.0000000000,
        ]
    elif koehli_method_form == "Aug08_mcnp_drf":
        p = [
            1.110773444917129,
            0.034319446894963,
            0.180046592985848,
            1.211393214064259,
            0.093433803170610,
            -1.877788035e-005,
            -0.00698637546803,
            5.0316941885e-005,
            0.0000000000,
        ]
    elif koehli_method_form == "Aug08_mcnp_ewin":
        p = [
            1.271225645585415,
            0.024790265564895,
            0.107603498535911,
            1.243101823658557,
            0.057146624195463,
            -1.93729201894976,
            -0.00866217333051,
            6.198559205414182,
            0.0000000000,
        ]
    elif koehli_method_form == "Aug12_uranos_drf":
        p = [
            1.042588152355816,
            0.024362250648228,
            0.222359434641456,
            1.791314246517330,
            0.197766380530824,
            -0.00053814104957,
            -0.00820189794785,
            6.6412111902e-005,
            0.0000000000,
        ]
    elif koehli_method_form == "Aug12_uranos_ewin":
        p = [
            1.209060105287452,
            0.021546879683024,
            0.129925023764294,
            1.872444149093526,
            0.128883139550384,
            -0.00047134595878,
            -0.01080226893400,
            8.8939419535e-005,
            0.0000000000,
        ]
    elif koehli_method_form == "Aug13_uranos_atmprof":
        p = [
            1.044276170094123,
            0.024099232055379,
            0.227317847739138,
            1.782905159416135,
            0.198949609723093,
            -0.00059182327737,
            -0.00897372356601,
            7.3282344356e-005,
            0.0000000000,
        ]
    elif koehli_method_form == "Aug13_uranos_atmprof2":
        p = [
            4.31237,
            0.020765,
            0.21020,
            1.87120,
            0.16341,
            -0.00052,
            -0.00225,
            0.000308,
            -1.9639e-8,
        ]

    N = (p[1] + p[2] * soil_moisture_total) / (soil_moisture_total + p[1]) * (
        p[0]
        + p[6] * abs_air_humidity
        + p[7] * abs_air_humidity**2
        + p[8] * abs_air_humidity**3 / soil_moisture_total
    ) + np.exp(-p[3] * soil_moisture_total) * (
        p[4] + p[5] * (abs_air_humidity + biomass / 5 * 1000)
    )

    return N * n0


"""
Examples
--------
import pandas

data = pandas.DataFrame()
data["N"] = [2000, 1900, 1800, 2100]
data["h"] = [1, 2, 3, 4]

data["sm"] = [
    convert_neutrons_to_soil_moisture_uts(
        neutron_count=N,
        n0=3000,
        air_humidity=h,
        bulk_density=1.23,
        lattice_water=0.01,
        water_equiv_soil_organic_carbon=0.01,
        method="Mar21_mcnp_drf",
    )
    for (N, h) in data[["N", "h"]]
]

data["sm"] = data.apply(
    lambda row: convert_neutrons_to_soil_moisture_uts(
        neutron_count=row["N"],
        n0=3000,
        air_humidity=row["h"],
        bulk_density=1.23,
        lattice_water=0.01,
        water_equiv_soil_organic_carbon=0.01,
        method="Mar21_mcnp_drf",
    ),
    axis=1,
)

data

soil_moisture_m3m3 = convert_neutrons_to_soil_moisture_uts(
    neutron_count=2000,
    n0=3000,
    air_humidity=5.0,
    bulk_density=1.23,
    lattice_water=0.0027,
    water_equiv_soil_organic_carbon=0.1,
    method="Mar21_mcnp_drf",
)
"""


def compute_n0_koehli_etal_2021(
    gravimetric_sm: float,
    abs_air_humidity: float,
    neutron_count: float,
    lattice_water=0.0,
    water_equiv_soil_organic_carbon=0.0,
    koehli_method_form: Literal[
        "Jan23_uranos",
        "Jan23_mcnpfull",
        "Mar12_atmprof",
        "Mar21_mcnp_drf",
        "Mar21_mcnp_ewin",
        "Mar21_uranos_drf",
        "Mar21_uranos_ewin",
        "Mar22_mcnp_drf_Jan",
        "Mar22_mcnp_ewin_gd",
        "Mar22_uranos_drf_gd",
        "Mar22_uranos_ewin_chi2",
        "Mar22_uranos_drf_h200m",
        "Aug08_mcnp_drf",
        "Aug08_mcnp_ewin",
        "Aug12_uranos_drf",
        "Aug12_uranos_ewin",
        "Aug13_uranos_atmprof",
        "Aug13_uranos_atmprof2",
    ] = "Mar21_uranos_drf",
    biomass=0,
):
    """
    Computes the n0 for the UTS-function following the method outlined
    in Köhli et al. 2021 (as needed by
    convert_neutrons_to_soil_moisture_uts() and
    convert_soil_moisture_to_neutrons_uts())

    https://doi.org/10.3389/frwa.2020.544847

    Parameters
    ----------
    gravimetric_sm : float
        gravimetric water content (g/cm3)
    abs_air_humidity : float
        absolute air humidity (g/cm3)
    bulk_density : float
        Dry soil bulk density of the soil in grams per cubic centimeter
        e.g. 1.4 (g/cm^3)
    neutron_count : float
        Neutron count in counts per hour (cph)
    lattice_water : float
        Lattice water - decimal percent e.g. 0.002
    water_equiv_soil_organic_carbon : float
        Water equivalent soil organic carbon - decimal percent e.g, 0.02
    method: str
        one of "Mar21_uranos_drf", "Aug13_uranos_atmprof", ...

    Example
    -------

    n0  = compute_n0_uts(soil_moisture=0.3, air_humidity=30,
    neutron_count=3000, lattice_water=0.0,
    water_equiv_soil_organic_carbon=0.0, bulk_density=1.2,
    method="Mar21_uranos_drf", bio=0)

    # check the determined n0, which should be close to n:
    convert_soil_moisture_to_neutrons_uts(
            sm=theta, air_humidity=30, n0=n0, offset=0.0,
            bulk_density=bulk_density, method="Mar21_uranos_drf",
            biomass=biomass)

    """
    from scipy.optimize import minimize_scalar

    off = lattice_water + water_equiv_soil_organic_carbon

    def obj_n0(n0_try):  # objective function to optimize for best n0
        neutron_estimate = (
            gravimetric_soil_moisture_to_neutrons_koehli_etal_2021(
                gravimetric_sm=gravimetric_sm,
                abs_air_humidity=abs_air_humidity,
                n0=n0_try,
                offset=off,
                koehli_method_form=koehli_method_form,
                biomass=biomass,
            )
        )
        error = np.abs(neutron_count - neutron_estimate)
        return np.mean(error)

    singleopt = minimize_scalar(obj_n0)  # optimize to find best n0
    n0 = singleopt.x  #

    return n0



import pandas as pd
from saqc import SaQC
import datetime
from neptoon.data_audit import log_key_step

# TODO Clean up these into one module


class TimeStampAligner:
    """
    Uses routines from SaQC to align the time stamps of the data to a
    common set. When data is read in it is added to an SaQC object which
    is stored as an internal feature. Data can then be aligned and
    converted back to a pd.DataFrame.

    Example
    -------
    >>> import pandas as pd
    >>> from neptoon.data_ingest_and_formatting.timestamp_alignment import (
    ...    TimeStampAligner
    ... )
    >>> data = {'value': [1, 2, 3, 4]}
    >>> index = pd.to_datetime(
    ...     [
    ...         "2021-01-01 00:04:00",
    ...         "2021-01-01 01:10:00",
    ...         "2021-01-01 02:05:00",
    ...         "2021-01-01 02:58:00",
    ...     ]
    ... )
    >>> df = pd.DataFrame(data, index=index)
    >>> # Initialize the TimeStampAligner
    >>> time_stamp_aligner = TimeStampAligner(df)
    >>> # Align timestamps
    >>> time_stamp_aligner.align_timestamps(method='nshift', freq='1H')
    >>> # Get the aligned dataframe
    >>> aligned_df = time_stamp_aligner.return_dataframe()
    >>> print(aligned_df)
    """

    def __init__(self, data_frame: pd.DataFrame):
        """
        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame containing time series data.
        """
        self._validate_timestamp_index(data_frame)
        self.data_frame = data_frame
        self.qc = SaQC(self.data_frame, scheme="simple")

    def _validate_timestamp_index(self, data_frame):
        """
        Checks that the index of the dataframe is timestamp (essential
        for aligning the time stamps and using SaQC)

        Parameters
        ----------
        data_frame : pd.DataFrame
            The data frame imported into the TimeStampAligner

        Raises
        ------
        ValueError
            If the index is not datetime type
        """
        if not pd.api.types.is_datetime64_any_dtype(data_frame.index):
            raise ValueError("The DataFrame index must be of datetime type")

    @staticmethod
    def timedelta_to_freq_str(time_delta: datetime.timedelta) -> str:
        """Convert a timedelta to a pandas frequency string."""
        total_seconds = time_delta.total_seconds()

        if total_seconds % (86400) == 0:  # Days (86400 = 24 * 60 * 60)
            return f"{int(total_seconds // 86400)}D"
        elif total_seconds % 3600 == 0:  # Hours
            return f"{int(total_seconds // 3600)}h"
        elif total_seconds % 60 == 0:  # Minutes
            return f"{int(total_seconds // 60)}min"
        else:  # Seconds
            return f"{int(total_seconds)}S"

    @log_key_step("method", "freq")
    def align_timestamps(self, freq: str = "1h", method: str = "time"):
        """
        Aligns the time stamp of the SaQC feature. Will automatically do
        this for all data columns. For more information on the values
        for method and freq see:

        https://rdm-software.pages.ufz.de/saqc/

        Parameters
        ----------
        method : str, optional
            Defaults to the nearest shift method to align time stamps.
            This means data is adjusted to the nearest time stamp
            without interpolation, by default "time".
        freq : str, optional
            The frequency of time stamps wanted, by default "1Hour"
        """

        if isinstance(freq, datetime.timedelta):
            freq = self.timedelta_to_freq_str(freq)

        self.qc = self.qc.align(
            field=self.data_frame.columns,
            freq=freq,
            method=method,
        )

    def return_dataframe(self):
        """
        Returns a pd.DataFrame from the SaQC object. Run this after
        alignment to return the aligned dataframe

        Returns
        -------
        df: pd.DataFrame
            DataFrame of time series data
        """
        df = self.qc.data.to_pandas()
        return df


class TimeStampAggregator:
    """
    Uses routines from SaQC to aggregate the data to a new sample rate.
    When data is read in it is added to an SaQC object which is stored
    as an internal feature. Data can then be aggregated and converted
    back to a pd.DataFrame.

    """

    def __init__(self, data_frame: pd.DataFrame):
        """
        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame containing time series data.
        """
        self._validate_timestamp_index(data_frame)
        self.data_frame = data_frame
        self.qc = SaQC(self.data_frame, scheme="simple")

    def _validate_timestamp_index(self, data_frame):
        """
        Checks that the index of the dataframe is timestamp (essential
        for aligning the time stamps and using SaQC)

        Parameters
        ----------
        data_frame : pd.DataFrame
            The data frame imported into the TimeStampAligner

        Raises
        ------
        ValueError
            If the index is not datetime type
        """
        if not pd.api.types.is_datetime64_any_dtype(data_frame.index):
            raise ValueError("The DataFrame index must be of datetime type")

    @staticmethod
    def timedelta_to_freq_str(time_delta: datetime.timedelta) -> str:
        """Convert a timedelta to a pandas frequency string."""
        total_seconds = time_delta.total_seconds()

        if total_seconds % (86400) == 0:  # Days (86400 = 24 * 60 * 60)
            return f"{int(total_seconds // 86400)}D"
        elif total_seconds % 3600 == 0:  # Hours
            return f"{int(total_seconds // 3600)}h"
        elif total_seconds % 60 == 0:  # Minutes
            return f"{int(total_seconds // 60)}min"
        else:  # Seconds
            return f"{int(total_seconds)}S"

    @log_key_step("method", "freq")
    def aggregate_data(
        self,
        freq: str = "1h",
        method: str = "bagg",
        func: str = "mean",
        max_na: float = 12.0,
    ):
        """
        Aggregates the data of the SaQC feature. Will automatically do
        this for all data columns. For more information on the values
        for method and freq see:

        https://rdm-software.pages.ufz.de/saqc/

        Parameters
        ----------
        method : str, optional
            Defaults to the nearest shift method to align time stamps.
            This means data is adjusted to the nearest time stamp
            without interpolation, by default "bagg".
        freq : str, optional
            The frequency of time stamps wanted, by default "1Hour"
        """

        if isinstance(freq, datetime.timedelta):
            freq = self.timedelta_to_freq_str(freq)

        self.qc = self.qc.resample(
            field=self.data_frame.columns,
            freq=freq,
            method=method,
            func=func,
            maxna=max_na,
        )

    def return_dataframe(self):
        """
        Returns a pd.DataFrame from the SaQC object. Run this after
        alignment to return the aligned dataframe

        Returns
        -------
        df: pd.DataFrame
            DataFrame of time series data
        """
        df = self.qc.data.to_pandas()
        return df



import pandas as pd
from enum import Enum
from abc import ABC, abstractmethod
from neptoon.logging import get_logger
from neptoon.columns import ColumnInfo

# read in the specific functions here
from neptoon.corrections import (
    incoming_intensity_correction,
    rc_correction_hawdon,
    McjannetDesilets2023,
    calc_absolute_humidity,
    calc_saturation_vapour_pressure,
    calc_actual_vapour_pressure,
    humidity_correction_rosolem2013,
    calc_pressure_correction_beta_coeff,
    calc_mean_pressure,
    calc_beta_coefficient,
    above_ground_biomass_correction_baatz2015,
    above_ground_biomass_correction_morris2024,
)
from neptoon.data_prep.conversions import AbsoluteHumidityCreator

core_logger = get_logger()


class CorrectionType(Enum):
    """
    The types of correction avaiable to implement.
    """

    INCOMING_INTENSITY = "incoming_intensity"
    ABOVE_GROUND_BIOMASS = "above_ground_biomass"
    PRESSURE = "pressure"
    HUMIDITY = "humidity"
    CUSTOM = "custom"


class CorrectionTheory(Enum):
    """
    The corrections theories for correcting influence on neutron signal
    beyond soil moisture
    """

    # Intensity
    ZREDA_2012 = "zreda_2012"
    HAWDON_2014 = "hawdon_2014"
    MCJANNET_DESILETS_2023 = "mcjannet_desilets_2023"
    # Atmospheric Water Vapour
    ROSOLEM_2013 = "rosolem_2013"
    # Pressure
    # Aboveground Biomass
    BAATZ_2015 = "baatz_2015"
    MORRIS_2024 = "morris_2024"


def is_column_missing_or_empty(data_frame, column_name):
    """
    Find whether a column is missing or empty in a dataframe. Useful for
    checking data before making calculations.

    Parameters
    ----------
    data_frame : pd.DataFrame
        _description_
    column_name : str
        Name of column to check for

    Returns
    -------
    bool
        True or False whether column is missing or empty
    """
    return (
        column_name not in data_frame.columns
        or data_frame[column_name].isnull().all()
    )


class Correction(ABC):
    """
    Abstract class for the Correction classes.

    All corrections have an apply method which takes a DataFrame as an
    argument. The return of the apply function should always be a
    DataFrame with the correction factor calculated and added as a
    column. The correction_factor_column_name should be set to the
    desired column name for the calculated correction factor.

    The CorrectionBuilder class will store the name of columns where
    correction factors are stored (when multiple corrections are
    undertaken). This enables the creation of the overall corrected
    neutron count column.
    """

    def __init__(
        self, correction_type: str, correction_factor_column_name: str
    ):
        self._correction_factor_column_name = correction_factor_column_name
        self.correction_type = correction_type

    @abstractmethod
    def apply(self, data_frame: pd.DataFrame):
        """
        The apply button should always take a dataframe as an input, do
        some logic, and return a dataframe with the additional columns
        calucalted during processing.

        Parameters
        ----------
        data_frame : pd.DataFrame
            The crns_data_frame
        """
        pass

    @property
    def correction_factor_column_name(self) -> str:
        if self._correction_factor_column_name is None:
            raise ValueError("correction_factor_column_name has not been set.")
        return self._correction_factor_column_name

    @correction_factor_column_name.setter
    def correction_factor_column_name(self, value: str):
        self._correction_factor_column_name = value

    def get_correction_factor_column_name(self):
        """
        Declare the name of the correction factor column
        """
        return self.correction_factor_column_name


class IncomingIntensityCorrectionZreda2012(Correction):
    """
    Corrects neutrons for incoming neutron intensity according to the
    original Zreda et al. (2012) equation.

    https://doi.org/10.5194/hess-16-4079-2012
    """

    def __init__(
        self,
        reference_incoming_neutron_value: str = str(
            ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE
        ),
        correction_type: str = CorrectionType.INCOMING_INTENSITY,
        correction_factor_column_name: str = str(
            ColumnInfo.Name.INTENSITY_CORRECTION
        ),
        incoming_neutron_column_name: str = str(
            ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY
        ),
    ):
        """
        Required attributes for creation.

        Parameters
        ----------
        reference_incoming_neutron_value : str
            name of the column holding the reference count of incoming
            neutron intensity at a point in time.
        correction_type : str
            correction type, by default CorrectionType.INCOMING_INTENSITY
        correction_factor_column_name : str, optional
            name of column corrections will be written to, by default
            ColumnInfo.Name.INTENSITY_CORRECTION
        incoming_neutron_column_name : str, optional
            name of column where incoming neutron intensity values are
            stored in the dataframe, by default
            "incoming_neutron_intensity"
        """
        super().__init__(
            correction_type=correction_type,
            correction_factor_column_name=correction_factor_column_name,
        )
        self.incoming_neutron_column_name = incoming_neutron_column_name
        self.reference_incoming_neutron_value = (
            reference_incoming_neutron_value
        )

    def apply(self, data_frame):
        """
        Applies the neutron correction

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with appropriate data

        Returns
        -------
        pd.DataFrame
            DataFrame now corrected
        """

        data_frame[self.correction_factor_column_name] = data_frame.apply(
            lambda row: incoming_intensity_correction(
                incoming_intensity=row[self.incoming_neutron_column_name],
                ref_incoming_intensity=row[
                    self.reference_incoming_neutron_value
                ],
                rc_scaling=1,
            ),
            axis=1,
        )

        return data_frame


class IncomingIntensityCorrectionHawdon2014(Correction):
    """
    Corrects for incoming neutron intensity according to the method
    outlined in Hawdon et al., (2014).

    https://doi.org/10.1002/2013WR015138
    """

    def __init__(
        self,
        ref_incoming_neutron_value: str = str(
            ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE
        ),
        site_cutoff_rigidity: str = str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY),
        correction_factor_column_name: str = str(
            ColumnInfo.Name.INTENSITY_CORRECTION
        ),
        incoming_neutron_column_name: str = str(
            ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY
        ),
        ref_monitor_cutoff_rigidity: str = str(
            ColumnInfo.Name.REFERENCE_MONITOR_CUTOFF_RIGIDITY
        ),
        rc_correction_factor: str = str(ColumnInfo.Name.RC_CORRECTION_FACTOR),
        correction_type: CorrectionType = CorrectionType.INCOMING_INTENSITY,
    ):
        """

        Parameters
        ----------
        ref_incoming_neutron_value : str, optional
            column name containing the reference incoming neutron value,
            by default str(
            ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE )
        site_cutoff_rigidity : str, optional
            column name containing the site cutoff rigidity, by default
            str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY)
        correction_factor_column_name : str, optional
            column name where correction factor will be written, by
            default str( ColumnInfo.Name.INTENSITY_CORRECTION )
        incoming_neutron_column_name : str, optional
            column name where incoming neutron intensity values are
            stored , by default str(
            ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY )
        ref_monitor_cutoff_rigidity : str, optional
            Name of the column the reference monitor cutoff rigidity is
            stored, by default str(
            ColumnInfo.Name.REFERENCE_MONITOR_CUTOFF_RIGIDITY )
        rc_correction_factor : str, optional
            name of the column the rc_correction will be written to, by
            default str( ColumnInfo.Name.RC_CORRECTION_FACTOR )
        correction_type : CorrectionType, optional
            The correction type as an Enum, by default
            CorrectionType.INCOMING_INTENSITY
        """
        super().__init__(
            correction_type=correction_type,
            correction_factor_column_name=correction_factor_column_name,
        )
        self.ref_incoming_neutron_value = ref_incoming_neutron_value
        self.site_cutoff_rigidity = site_cutoff_rigidity
        self.ref_monitor_cutoff_rigidity = ref_monitor_cutoff_rigidity
        self.rc_correction_factor = rc_correction_factor
        self.incoming_neutron_column_name = incoming_neutron_column_name

    def _check_required_columns(self, data_frame):
        """
        Checks that the required columns are available.

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with the data

        Raises
        ------
        ValueError
            Raises when required columns are missing
        """
        required_columns = [
            self.incoming_neutron_column_name,
            self.ref_incoming_neutron_value,
            self.site_cutoff_rigidity,
        ]
        missing_columns = [
            col
            for col in required_columns
            if is_column_missing_or_empty(data_frame, col)
        ]
        if missing_columns:
            raise ValueError(
                f"Required columns are missing or empty: {', '.join(missing_columns)}"
            )

    def _calc_rc_scale_param(self, data_frame):
        """
        Calculates the correction to account for difference in site and
        reference monitor cutoff rigidity.

        Parameters
        ----------
        data_frame : pd.DataFrame
            The DataFrame with the data
        """
        data_frame[self.rc_correction_factor] = data_frame.apply(
            lambda row: rc_correction_hawdon(
                site_cutoff_rigidity=row[self.site_cutoff_rigidity],
                ref_monitor_cutoff_rigidity=row[
                    self.ref_monitor_cutoff_rigidity
                ],
            ),
            axis=1,
        )
        return data_frame

    def apply(self, data_frame):
        """
        Applies the correction factor returning a dataframe with the
        calculated values in it.

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame

        Returns
        -------
        pd.DataFrame
            DataFrame with the correction values attached
        """
        self._check_required_columns(data_frame=data_frame)
        data_frame = self._calc_rc_scale_param(data_frame=data_frame)

        data_frame[self.correction_factor_column_name] = data_frame.apply(
            lambda row: incoming_intensity_correction(
                incoming_intensity=row[self.incoming_neutron_column_name],
                ref_incoming_intensity=row[self.ref_incoming_neutron_value],
                rc_scaling=row[self.rc_correction_factor],
            ),
            axis=1,
        )
        return data_frame


class IncomingIntensityCorrectionMcJannetDesilets2023(Correction):
    """
    Corrects for incoming neutron intensity according to the method
    outlined in McJannet and Desilets., (2023).

    https://doi.org/10.1029/2022WR033889
    """

    def __init__(
        self,
        ref_incoming_neutron_value: str = str(
            ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE
        ),
        site_cutoff_rigidity: str = str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY),
        correction_factor_column_name: str = str(
            ColumnInfo.Name.INTENSITY_CORRECTION
        ),
        incoming_neutron_column_name: str = str(
            ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY
        ),
        latitude: str = str(ColumnInfo.Name.LATITUDE),
        elevation: str = str(ColumnInfo.Name.ELEVATION),
        rc_correction_factor: str = str(ColumnInfo.Name.RC_CORRECTION_FACTOR),
        correction_type: CorrectionType = CorrectionType.INCOMING_INTENSITY,
    ):
        """

        Parameters
        ----------
        ref_incoming_neutron_value : str, optional
            column name containing the reference incoming neutron value,
            by default str(
            ColumnInfo.Name.REFERENCE_INCOMING_NEUTRON_VALUE )
        site_cutoff_rigidity : str, optional
            column name containing the site cutoff rigidity, by default
            str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY)
        correction_factor_column_name : str, optional
            column name where correction factor will be written, by
            default str( ColumnInfo.Name.INTENSITY_CORRECTION )
        incoming_neutron_column_name : str, optional
            column name where incoming neutron intensity values are
            stored, by default str(
            ColumnInfo.Name.INCOMING_NEUTRON_INTENSITY )
        latitude : str, optional
            column name where latitude values are stored, by default
            str(ColumnInfo.Name.LATITUDE)
        elevation : str, optional
            column name where elevation values are stored, by default
            str(ColumnInfo.Name.ELEVATION)
        rc_correction_factor : str, optional
            name of the column the rc_correction will be written to, by
            default str(ColumnInfo.Name.RC_CORRECTION_FACTOR)
        correction_type : CorrectionType, optional
            the correction type as an Enum, by default
            CorrectionType.INCOMING_INTENSITY
        """
        super().__init__(
            correction_type=correction_type,
            correction_factor_column_name=correction_factor_column_name,
        )
        self.ref_incoming_neutron_value = ref_incoming_neutron_value
        self.site_cutoff_rigidity = site_cutoff_rigidity
        self.latitude = latitude
        self.elevation = elevation
        self.rc_correction_factor = rc_correction_factor
        self.incoming_neutron_column_name = incoming_neutron_column_name

    def _check_required_columns(self, data_frame):
        """
        Checks that the required columns are available.

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame to check

        Raises
        ------
        ValueError
            Error when needed columns are missing.
        """
        required_columns = [
            self.incoming_neutron_column_name,
            self.ref_incoming_neutron_value,
            self.site_cutoff_rigidity,
            self.latitude,
            self.elevation,
        ]

        missing_columns = [
            col
            for col in required_columns
            if is_column_missing_or_empty(data_frame, col)
        ]
        if missing_columns:
            raise ValueError(
                f"Required columns are missing or empty: {', '.join(missing_columns)}"
            )

    def _check_reference_monitor_is_jung(self, data_frame):
        """
        The reference station in the mcjannet desillets formulation is
        Jungfraujoch. If different reference monitor is supplied, and
        therefore a different reference point, an error is raised.

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with data
        """
        station_supplied = data_frame[
            str(ColumnInfo.Name.NMDB_REFERENCE_STATION)
        ].iloc[0]
        if station_supplied != "JUNG":
            message = (
                "If using the McjannetDesilets2023 method for incoming intensity correction "
                "only 'JUNG' (Jungfraujoch) can be given as a reference station. \n"
                f"{station_supplied} was given"
            )
            core_logger.error(message)
            raise ValueError(message)

    def _calc_rc_scale_param(self, data_frame):
        """
        Calculates the correction to account for difference in site and
        reference monitor cutoff rigidity.

        Parameters
        ----------
        data_frame : pd.DataFrame
            The DataFrame with the data
        """
        data_frame[self.rc_correction_factor] = data_frame.apply(
            lambda row: McjannetDesilets2023.tau(
                latitude=row[self.latitude],
                elevation=row[self.elevation],
                cut_off_rigidity=row[self.site_cutoff_rigidity],
            ),
            axis=1,
        )
        return data_frame

    def apply(self, data_frame):
        self._check_required_columns(data_frame=data_frame)
        self._check_reference_monitor_is_jung(data_frame=data_frame)
        data_frame = self._calc_rc_scale_param(data_frame=data_frame)

        data_frame[self.correction_factor_column_name] = data_frame.apply(
            lambda row: incoming_intensity_correction(
                incoming_intensity=row[self.incoming_neutron_column_name],
                ref_incoming_intensity=row[self.ref_incoming_neutron_value],
                rc_scaling=row[self.rc_correction_factor],
            ),
            axis=1,
        )
        return data_frame


class HumidityCorrectionRosolem2013(Correction):
    """
    Corrects neutrons for humidity according to the
     Rosolem et al. (2013) equation.

    https://doi.org/10.1175/JHM-D-12-0120.1
    """

    def __init__(
        self,
        reference_absolute_humidity_value: float = 0,
        correction_type: str = CorrectionType.HUMIDITY,
        correction_factor_column_name: str = str(
            ColumnInfo.Name.HUMIDITY_CORRECTION
        ),
        sat_vapour_pressure_column_name: str = str(
            ColumnInfo.Name.SATURATION_VAPOUR_PRESSURE
        ),
        air_temperature_column_name: str = str(
            ColumnInfo.Name.AIR_TEMPERATURE
        ),
        actual_vapour_pressure_column_name: str = str(
            ColumnInfo.Name.ACTUAL_VAPOUR_PRESSURE
        ),
        absolute_humidity_column_name: str = str(
            ColumnInfo.Name.ABSOLUTE_HUMIDITY
        ),
        relative_humidity_column_name: str = str(
            ColumnInfo.Name.AIR_RELATIVE_HUMIDITY
        ),
    ):
        """
        Required attributes for creation.

        Parameters
        ----------
        reference_incoming_neutron_value : float
            reference count of incoming neutron intensity at a point in
            time.
        correction_type : str, optional
            correction type, by default "intensity"
        correction_factor_column_name : str, optional
            name of column corrections will be written to, by default
            "correction_for_intensity"
        incoming_neutron_column_name : str, optional
            name of column where incoming neutron intensity values are
            stored in the dataframe, by default
            "incoming_neutron_intensity"
        """
        super().__init__(
            correction_type=correction_type,
            correction_factor_column_name=correction_factor_column_name,
        )
        self.sat_vapour_pressure_column_name = sat_vapour_pressure_column_name
        self.reference_absolute_humidity_value = (
            reference_absolute_humidity_value
        )
        self.air_temperature_column_name = air_temperature_column_name
        self.absolute_humidity_column_name = absolute_humidity_column_name
        self.actual_vapour_pressure_column_name = (
            actual_vapour_pressure_column_name
        )
        self.relative_humidity_column_name = relative_humidity_column_name

    def _check_if_abs_hum_exists(self, data_frame):
        """
        Checks if absolute humidity column exits

        Parameters
        ----------
        data_frame : pd.DataFrame
            Data frame with sensor data

        Returns
        -------
        Bool
            True or False if abs hum col exists.
        """
        if self.absolute_humidity_column_name in data_frame.columns:
            return True
        else:
            return False

    def _create_abs_hum_data(self, data_frame):
        """
        Creates absolute humidity data

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with CRNS data

        Returns
        -------
        pd.DataFrame
            With additional columns such as absolute humidity
        """
        abs_hum_creator = AbsoluteHumidityCreator(data_frame=data_frame)
        data_frame = abs_hum_creator.check_and_return_abs_hum_column()
        return data_frame

    def apply(self, data_frame):
        """
        Applies the neutron correction

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with appropriate data

        Returns
        -------
        pd.DataFrame
            DataFrame now corrected
        """

        abs_hum_exists = self._check_if_abs_hum_exists(data_frame=data_frame)
        if not abs_hum_exists:
            data_frame = self._create_abs_hum_data(data_frame=data_frame)

        data_frame[self.correction_factor_column_name] = data_frame.apply(
            lambda row: humidity_correction_rosolem2013(
                row[self.absolute_humidity_column_name],
                self.reference_absolute_humidity_value,
            ),
            axis=1,
        )

        return data_frame


class PressureCorrectionZreda2012(Correction):
    """
    Corrects neutrons for changes in atmospheric pressure according to
    the original Zreda et al. (2012) equation.

    https://doi.org/10.5194/hess-16-4079-2012
    """

    def __init__(
        self,
        site_elevation: str = str(ColumnInfo.Name.ELEVATION),
        reference_pressure_value: str = str(ColumnInfo.Name.MEAN_PRESSURE),
        correction_type: str = CorrectionType.PRESSURE,
        correction_factor_column_name: str = str(
            ColumnInfo.Name.PRESSURE_CORRECTION
        ),
        beta_coefficient: str = str(ColumnInfo.Name.BETA_COEFFICIENT),
        latitude: str = str(ColumnInfo.Name.LATITUDE),
        site_cutoff_rigidity: str = str(ColumnInfo.Name.SITE_CUTOFF_RIGIDITY),
    ):
        """
        Required attributes for creation.

        Parameters
        ----------
        site_elevation : float, optional
            site elevation in m, by default None
        reference_pressure_value : float, optional
            reference pressure for correction (recommended to be average
            site pressure). - hPa , by default None
        correction_type : str, optional
            correction type, by default CorrectionType.PRESSURE
        correction_factor_column_name : str, optional
            Name of column to store correction factors, by default str(
            ColumnInfo.Name.PRESSURE_CORRECTION )
        beta_coefficient : float, optional
            beta_coefficient for processing, by default None
        latitude : float, optional
            latitude of site in degrees, by default None
        site_cutoff_rigidity : _type_, optional
            cut-off rigidity at the site, by default None
        """
        super().__init__(
            correction_type=correction_type,
            correction_factor_column_name=correction_factor_column_name,
        )
        self.reference_pressure_value = reference_pressure_value
        self.beta_coefficient = beta_coefficient
        self.site_elevation = site_elevation
        self.latitude = latitude
        self.site_cutoff_rigidity = site_cutoff_rigidity

    def _prepare_for_correction(self, data_frame):
        """
        Prepare to correction process. Check to see if reference
        pressure needs calculating and then checks for coefficients
        given in site information. If no coefficient given it will
        calculate the beta_coefficient.
        """

        self._ensure_reference_pressure_available(data_frame)
        self._check_coefficient_available(data_frame)

    def _ensure_reference_pressure_available(self, data_frame):
        """
        Checks for reference pressure.

        NOTE: Important to note that changing reference pressure from
        the value used during calibration will impact the results.
        If reference pressure is changed for processing the site must be
        re-calibrated so that the N0 has the same reference pressure.

        Raises
        ------
        ValueError
            If no reference pressure and no elevation it cannot work.
            Raises error.
        """
        column_name_press = self.reference_pressure_value
        column_name_elev = self.site_elevation
        if is_column_missing_or_empty(data_frame, column_name_press):
            if is_column_missing_or_empty(data_frame, column_name_elev):
                message = (
                    "You must supply a reference pressure or a site elevation"
                )
                core_logger.error(message)
                raise ValueError(message)
            message = (
                "No reference pressure value given. Calculating average pressure "
                "using elevation information and using this value"
            )
            core_logger.info(message)
            data_frame[self.reference_pressure_value] = data_frame.apply(
                lambda row: calc_mean_pressure(row[self.site_elevation]),
                axis=1,
            )

    def _check_coefficient_available(self, data_frame):
        """
        Checks for coefficient availability. If not available it will
        calculate it using latitude, site elevation and site cutoff
        rigidity.
        """
        column_name_beta = self.beta_coefficient

        if is_column_missing_or_empty(data_frame, column_name_beta):
            message = (
                "No coefficient given for pressure correction. "
                "Calculating beta coefficient."
            )
            core_logger.info(message)
            data_frame[self.beta_coefficient] = data_frame.apply(
                lambda row: calc_beta_coefficient(
                    row[self.reference_pressure_value],
                    row[self.latitude],
                    row[self.site_elevation],
                    row[self.site_cutoff_rigidity],
                ),
                axis=1,
            )

    def apply(self, data_frame):
        """
        Applies the neutron correction

        Parameters
        ----------
        data_frame : pd.DataFrame
            DataFrame with appropriate data

        Returns
        -------
        pd.DataFrame
            DataFrame now corrected
        """

        # TODO validation here

        if not is_column_missing_or_empty(
            data_frame, self.correction_factor_column_name
        ):
            message = (
                "The correction already appears in the data_frame as"
                f"'{self.correction_factor_column_name}'. Skipping correction to prevent "
                "unwanted overwrites. "
            )
            core_logger.info(message)
            return data_frame
        else:
            self._prepare_for_correction(data_frame)
            data_frame[self.correction_factor_column_name] = data_frame.apply(
                lambda row: calc_pressure_correction_beta_coeff(
                    row[str(ColumnInfo.Name.AIR_PRESSURE)],
                    row[self.reference_pressure_value],
                    row[self.beta_coefficient],
                ),
                axis=1,
            )
            return data_frame


class AboveGroundBiomassCorrectionBaatz2015(Correction):

    def __init__(
        self,
        correction_type: CorrectionType,
        correction_factor_column_name: str,
        above_ground_biomass_column_name: str,
    ):
        super().__init__(
            correction_type=correction_type,
            correction_factor_column_name=correction_factor_column_name,
        )
        self.above_ground_biomass_column_name = (
            above_ground_biomass_column_name
        )

    def apply(self, data_frame):
        print(
            f"{self.above_ground_biomass_column_name} needs to be in units of "
            "kg m^2."
        )
        data_frame[self.correction_factor_column_name] = data_frame.apply(
            lambda row: above_ground_biomass_correction_baatz2015(
                above_ground_biomass=row[self.above_ground_biomass_column_name]
            ),
            axis=1,
        )

        return data_frame


class AboveGroundBiomassCorrectionMorris2024(Correction):
    """_summary_



    https://doi.org/10.3390/s24134094
    """

    def __init__(
        self,
        correction_type: CorrectionType,
        correction_factor_column_name: str,
        above_ground_biomass_column_name: str,
    ):
        super().__init__(
            correction_type=correction_type,
            correction_factor_column_name=correction_factor_column_name,
        )
        self.above_ground_biomass_column_name = (
            above_ground_biomass_column_name
        )

    def apply(self, data_frame):
        print(
            f"{self.above_ground_biomass_column_name} needs to be in units of "
            "mm (i.e., Biomass Water Equivalant)."
        )
        data_frame[self.correction_factor_column_name] = data_frame.apply(
            lambda row: above_ground_biomass_correction_morris2024(
                above_ground_biomass=row[self.above_ground_biomass_column_name]
            ),
            axis=1,
        )

        return data_frame
